"""Schema for various output tables produced by hpmcm"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from .cluster import ShearClusterData
from .object import ShearObjectData
from .table import TableColumnInfo, TableInterface

if TYPE_CHECKING:
    from .cell import CellData


SHEAR_NAMES = ["ns", "2p", "2m", "1p", "1m"]


class ObjectAssocTable(TableInterface):
    """Interface of table with associations between objects and sources"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        objectId=TableColumnInfo(int, "Unique Object ID"),
        clusterId=TableColumnInfo(int, "Parent Cluster Unique ID"),
        sourceId=TableColumnInfo(int, "Source id in input catalog"),
        sourceIdx=TableColumnInfo(int, "Source index in input catalog"),
        catalogId=TableColumnInfo(int, "Associated catalog ID"),
        distance=TableColumnInfo(float, "Distance from sources to object centroid"),
        cellIdx=TableColumnInfo(int, "Index of associated cell"),
    )

    @staticmethod
    def buildFromCellData(cellData: CellData) -> ObjectAssocTable:
        """Create object association table

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Object Association table
        """
        clusterIds = []
        objectIds = []
        sourceIds = []
        sourceIdxs = []
        catIdxs = []
        distancesList: list[np.ndarray] = []

        for obj in cellData.objectDict.values():
            clusterIds.append(
                np.full((obj.nSrc), obj.parentCluster.iCluster, dtype=int)
            )
            objectIds.append(np.full((obj.nSrc), obj.objectId, dtype=int))
            sourceIds.append(obj.sourceIds())
            sourceIdxs.append(obj.sourceIdxs())
            catIdxs.append(obj.catIndices)
            assert obj.dist2.size
            distancesList.append(obj.dist2)
        if not distancesList:
            return ObjectAssocTable(
                objectId=np.array([], int),
                clusterId=np.array([], int),
                sourceId=np.array([], int),
                sourceIdx=np.array([], int),
                catalogId=np.array([], int),
                distance=np.array([], float),
                cellIdx=np.array([], int),
            )
        distances = np.hstack(distancesList)
        distances = cellData.matcher.pixToArcsec() * np.sqrt(distances)
        return ObjectAssocTable(
            objectId=np.hstack(objectIds),
            clusterId=np.hstack(clusterIds),
            sourceId=np.hstack(sourceIds),
            sourceIdx=np.hstack(sourceIdxs),
            catalogId=np.hstack(catIdxs),
            distance=distances,
            cellIdx=np.repeat(cellData.idx, len(distances)).astype(int),
        )


class ObjectStatsTable(TableInterface):
    """Interface of table of object statistics"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        objectId=TableColumnInfo(int, "Unique Object ID"),
        clusterId=TableColumnInfo(int, "Parent Cluster Unique ID"),
        nUnique=TableColumnInfo(int, "Number of unique catalogs represented"),
        nSrc=TableColumnInfo(int, "Number of sources"),
        distRms=TableColumnInfo(
            float, "RMS of distance from sources to object centroid"
        ),
        ra=TableColumnInfo(float, "RA of object centroid"),
        dec=TableColumnInfo(float, "DEC of object centroid"),
        xCent=TableColumnInfo(float, "X-value of cluster centroid in WCS pixels"),
        yCent=TableColumnInfo(float, "Y-value of cluster centroid in WCS pixels"),
        SNR=TableColumnInfo(float, "Mean signal-to-noise ratio"),
        SNRRms=TableColumnInfo(float, "RMS signal-to-noise ratio"),
        cellIdx=TableColumnInfo(int, "Index of associated cell"),
        hasRefCat=TableColumnInfo(bool, "Has source from the reference catalog"),
    )

    @staticmethod
    def buildFromCellData(cellData: CellData) -> ObjectStatsTable:
        """Create object stats table

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Object stats table
        """
        nObj = cellData.nObjects
        clusterIds = np.zeros((nObj), dtype=int)
        objectIds = np.zeros((nObj), dtype=int)
        nSrcs = np.zeros((nObj), dtype=int)
        nUniques = np.zeros((nObj), dtype=int)
        distRms = np.zeros((nObj), dtype=float)
        xCents = np.zeros((nObj), dtype=float)
        yCents = np.zeros((nObj), dtype=float)
        SNRs = np.zeros((nObj), dtype=float)
        SNRRms = np.zeros((nObj), dtype=float)
        hasRefCat = np.zeros((nObj), dtype=bool)

        for idx, obj in enumerate(cellData.objectDict.values()):
            clusterIds[idx] = obj.parentCluster.iCluster
            objectIds[idx] = obj.objectId
            nSrcs[idx] = obj.nSrc
            nUniques[idx] = obj.nUnique
            distRms[idx] = obj.rmsDist
            xCents[idx] = obj.xCent
            yCents[idx] = obj.yCent
            assert obj.data is not None
            sumSNR = obj.data.SNR.sum()
            xCents[idx] = np.sum(obj.data.SNR * obj.data.xCell) / sumSNR
            yCents[idx] = np.sum(obj.data.SNR * obj.data.yCell) / sumSNR
            SNRs[idx] = obj.snrMean
            SNRRms[idx] = obj.snrRms
            hasRefCat[idx] = obj.hasRefCatalog()

        ra, dec = cellData.getRaDec(xCents, yCents)
        distRms *= cellData.matcher.pixToArcsec()

        return ObjectStatsTable(
            clusterId=clusterIds,
            objectId=objectIds,
            nUnique=nUniques,
            nSrc=nSrcs,
            distRms=distRms,
            ra=ra,
            dec=dec,
            xCent=xCents,
            yCent=yCents,
            SNR=SNRs,
            SNRRms=SNRRms,
            cellIdx=np.repeat(cellData.idx, len(distRms)).astype(int),
            hasRefCat=hasRefCat,
        )


class ClusterAssocTable(TableInterface):
    """Interface of table with associations between clusters and sources"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        clusterId=TableColumnInfo(int, "Unique cluster ID"),
        sourceId=TableColumnInfo(int, "Source id in input catalog"),
        sourceIdx=TableColumnInfo(int, "Source index in input catalog"),
        catalogId=TableColumnInfo(int, "Associated catalog ID"),
        distance=TableColumnInfo(float, "Distance from sources to cluster centroid"),
        cellIdx=TableColumnInfo(int, "Index of associated cell"),
    )

    @staticmethod
    def buildFromCellData(cellData: CellData) -> ClusterAssocTable:
        """Create object association table

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Cluster Association table
        """
        clusterIds = []
        sourceIds = []
        sourceIdxs = []
        catIdxs = []
        distancesList: list[np.ndarray] = []
        for cluster in cellData.clusterDict.values():
            clusterIds.append(np.full((cluster.nSrc), cluster.iCluster, dtype=int))
            sourceIds.append(cluster.srcIds)
            sourceIdxs.append(cluster.srcIdxs)
            catIdxs.append(cluster.catIndices)
            assert cluster.dist2.size
            distancesList.append(cluster.dist2)
        if not distancesList:
            return ClusterAssocTable(
                distance=np.array([], float),
                sourceId=np.array([], int),
                sourceIdx=np.array([], int),
                catalogId=np.array([], int),
                clusterId=np.array([], int),
                cellIdx=np.array([], int),
            )
        distances = np.hstack(distancesList)
        distances = cellData.matcher.pixToArcsec() * np.sqrt(distances)
        return ClusterAssocTable(
            clusterId=np.hstack(clusterIds),
            sourceId=np.hstack(sourceIds),
            sourceIdx=np.hstack(sourceIdxs),
            catalogId=np.hstack(catIdxs),
            distance=distances,
            cellIdx=np.repeat(cellData.idx, len(distances)).astype(int),
        )


class ClusterStatsTable(TableInterface):
    """Interface of table of cluster statistics"""

    _schema = TableInterface._schema.copy()
    _schema.update(
        clusterId=TableColumnInfo(int, "Parent Cluster Unique ID"),
        nObject=TableColumnInfo(int, "Number of objects in cluster"),
        nUnique=TableColumnInfo(int, "Number of unique catalogs represented"),
        nSrc=TableColumnInfo(int, "Number of sources"),
        distRms=TableColumnInfo(
            float, "RMS of distance from sources to object centroid"
        ),
        ra=TableColumnInfo(float, "RA of object centroid"),
        dec=TableColumnInfo(float, "DEC of object centroid"),
        xCent=TableColumnInfo(float, "X-value of cluster centroid in WCS pixels"),
        yCent=TableColumnInfo(float, "Y-value of cluster centroid in WCS pixels"),
        SNR=TableColumnInfo(float, "Mean signal-to-noise ratio"),
        SNRRms=TableColumnInfo(float, "RMS signal-to-noise ratio"),
        cellIdx=TableColumnInfo(int, "Index of associated cell"),
        hasRefCat=TableColumnInfo(bool, "Has source from reference catalog"),
    )

    @staticmethod
    def buildFromCellData(cellData: CellData) -> ClusterStatsTable:
        """Create object stats table

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Object stats table
        """
        nClust = cellData.nClusters
        clusterIds = np.zeros((nClust), dtype=int)
        nSrcs = np.zeros((nClust), dtype=int)
        nUniques = np.zeros((nClust), dtype=int)
        nObjects = np.zeros((nClust), dtype=int)
        nUniques = np.zeros((nClust), dtype=int)
        distRms = np.zeros((nClust), dtype=float)
        xCents = np.zeros((nClust), dtype=float)
        yCents = np.zeros((nClust), dtype=float)
        SNRs = np.zeros((nClust), dtype=float)
        SNRRms = np.zeros((nClust), dtype=float)
        hasRefCat = np.zeros((nClust), dtype=bool)

        for idx, cluster in enumerate(cellData.clusterDict.values()):
            clusterIds[idx] = cluster.iCluster
            nSrcs[idx] = cluster.nSrc
            nUniques[idx] = cluster.nUnique
            nObjects[idx] = len(cluster.objects)
            nUniques[idx] = cluster.nUnique
            distRms[idx] = cluster.rmsDist
            assert cluster.data is not None
            sumSNR = cluster.data.SNR.sum()
            xCents[idx] = np.sum(cluster.data.SNR * cluster.data.xCell) / sumSNR
            yCents[idx] = np.sum(cluster.data.SNR * cluster.data.yCell) / sumSNR
            SNRs[idx] = cluster.snrMean
            SNRRms[idx] = cluster.snrRms
            hasRefCat[idx] = cluster.hasRefCatalog()

        ra, dec = cellData.getRaDec(xCents, yCents)
        distRms *= cellData.matcher.pixToArcsec()

        return ClusterStatsTable(
            clusterId=clusterIds,
            nSrc=nSrcs,
            nObject=nObjects,
            nUnique=nUniques,
            distRms=distRms,
            ra=ra,
            dec=dec,
            xCent=xCents,
            yCent=yCents,
            SNR=SNRs,
            SNRRms=SNRRms,
            cellIdx=np.repeat(cellData.idx, len(distRms)).astype(int),
            hasRefCat=hasRefCat,
        )


class ShearTable(TableInterface):
    """Interface of table with shear information"""

    _schema = TableInterface._schema.copy()
    _schema["good"] = TableColumnInfo(bool, "Has unique match")
    for _name in SHEAR_NAMES:
        _schema[f"n_{_name}"] = TableColumnInfo(
            float, f"number of sourcrs from catalog {_name}"
        )
        for _i in [1, 2]:
            _schema[f"g_{_i}_{_name}"] = TableColumnInfo(
                float, f"g {_i} for catalog {_name}"
            )
    for _i in [1, 2]:
        for _j in [1, 2]:
            _schema[f"delta_g_{_i}_{_j}"] = TableColumnInfo(
                float, f"delta g {_i} for {_j}p - {_j}m"
            )

    @classmethod
    def buildObjectShearStats(cls, cellData: CellData) -> ShearTable:
        """Create shear stats table for objects in a cell

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Shear stats table
        """
        nObj = cellData.nObjects
        outDict = ShearTable.emtpyNumpyDict(nObj)
        for idx, obj in enumerate(cellData.objectDict.values()):
            assert isinstance(obj, ShearObjectData)
            objStats = obj.shearStats()
            for key, val in objStats.items():
                outDict[key][idx] = val
        return ShearTable(**outDict)

    @classmethod
    def buildClusterShearStats(cls, cellData: CellData) -> ShearTable:
        """Create shear stats table for clusters in a cell

        Parameters
        ----------
        cellData:
            Cell we are making table for

        Returns
        -------
        Shear stats table
        """
        nClusters = cellData.nClusters
        outDict = ShearTable.emtpyNumpyDict(nClusters)
        for idx, obj in enumerate(cellData.clusterDict.values()):
            assert isinstance(obj, ShearClusterData)
            objStats = obj.shearStats()
            for key, val in objStats.items():
                outDict[key][idx] = val
        return ShearTable(**outDict)



import glob
import os

import numpy as np
import hpmcm

DATADIR = "examples/test_data"  # Input data directory
SHEAR_ST = "0p01"  # Applied shear as a string
TRACT = 10463  # which tract to study

REF_DIR = (37.9, 7.0)  # RA, DEC in deg of center of match region
REGION_SIZE = (0.375, 0.375)  # Size of match region in degrees
PIXEL_SIZE = 0.5 / 3600.0  # Size of pixels used in matching
PIXEL_R2CUT = 4.0  # Cut at distance**2 = 4 pixels


def testWCSMatch(setup_data: int) -> None:
    """Run WcsMatch on a large part of a tract"""

    assert setup_data == 0

    # get the data
    
    source_tablesfiles = sorted(
        glob.glob(os.path.join(DATADIR, f"shear_*_{SHEAR_ST}_cleaned_{TRACT}_ns.pq"))
    )
    source_tablesfiles.append(os.path.join(DATADIR, f"object_{TRACT}.pq"))
    source_tablesfiles.reverse()
    source_tablesfiles = [source_tablesfiles[0], source_tablesfiles[1]]
    catalog_ids = np.arange(len(source_tablesfiles))

    # Create matcher
    matcher = hpmcm.WcsMatch.create(
        REF_DIR, REGION_SIZE, pixelSize=PIXEL_SIZE, pixelR2Cut=PIXEL_R2CUT
    )

    # Reduce the input data
    matcher.reduceData(source_tablesfiles, catalog_ids)

    # Make sure it got the right number of cells
    assert matcher.nCell[0] == 3
    assert matcher.nCell[1] == 3

    # Define the range of cells to run over
    xRange = range(1, 2)
    yRange = range(1, 2)

    # Run the analysis
    matcher.analysisLoop(xRange, yRange)

    # Extract some data
    stats = matcher.extractStats()
    assert stats is not None

    # Test the classification codes
    objLists = hpmcm.classify.classifyObjects(matcher, SNRCut=10.0)
    hpmcm.classify.printObjectTypes(objLists)

    # Test the classification codes
    odict = hpmcm.classify.matchObjectsAgainstRef(matcher, snrCut=10.0)
    hpmcm.classify.printObjectMatchTypes(odict)

    # Make sure the match efficiency is high
    n_good = len(objLists["ideal"])
    bad_list = [
        "edge_mixed",
        "edge_missing",
        "edge_extra",
        "orphan",
        "missing",
        "two_missing",
        "many_missing",
        "extra",
        "caught",
    ]
    n_bad = np.sum([len(objLists[x]) for x in bad_list])
    effic = n_good / (n_good + n_bad)

    assert effic > 0.85

    # Get a particular cell and rerun the analysis to test visualization
    # and classification functions
    cell = matcher.cellDict[matcher.getCellIdx(1, 1)]
    _od = cell.analyze(None, 4)
    _cluster = list(cell.clusterDict.values())[0]



from __future__ import annotations

from typing import Any

import numpy as np
import pandas

from . import input_tables, output_tables
from .cell import CellData, ShearCellData
from .match import Match


class ShearMatch(Match):
    """Class to do N-way matching for shear calibration.

    Uses pre-assigned pixel locations from cell-based coadd WCS.

    Expects 5 input catalogs: a reference catalog and 4 counterfactual
    shear catalogs.

    Attributes
    ----------
    pixelMatchScale: int
        Number of cells to merge in original counts map

    catType: str
        Shear catalog type

    deshear: float | None
        Deshearing parameter, -1*applied shear.  None -> dshearing is not done.

    Notes
    -----
    This expectes a list of parquet files with pandas DataFrames
    that contain the following columns.

    +-------------+---------------------------------------------------------------+
    | Column name | Description                                                   |
    +=============+===============================================================+
    | id          | source ID                                                     |
    +-------------+---------------------------------------------------------------+
    | tract       | Tract being matched                                           |
    +-------------+---------------------------------------------------------------+
    | xCellCoadd  | X-postion in cell-based coadd used for metadetect             |
    +-------------+---------------------------------------------------------------+
    | yCellCoadd  | Y-postion in cell-based coadd used for metadetect             |
    +-------------+---------------------------------------------------------------+
    | SNR         | Signal-to-Noise of source, used for filtering and centroiding |
    +-------------+---------------------------------------------------------------+
    | cellIdxX    | Cell x-index within Tract                                     |
    +-------------+---------------------------------------------------------------+
    | cellIdxY    | Cell y-index within Tract                                     |
    +-------------+---------------------------------------------------------------+
    | g_1         | Shear g1 component                                            |
    +-------------+---------------------------------------------------------------+
    | g_2         | Shear g1 component                                            |
    +-------------+---------------------------------------------------------------+

    (see :py:class:`hpmcm.input_tables.ShearCoaddSourceTable`)


    These parquet files can be generated from files with the following
    columns using the ShearMatch.splitByTypeAndClean() function.

    +--------------------------------+---------------------------------------+
    | Column name                    | Description                           |
    +================================+=======================================+
    | id                             | source ID                             |
    +--------------------------------+---------------------------------------+
    | shear_type                     | one of "ns", "1p", "1m", "2p" "2m"    |
    +--------------------------------+---------------------------------------+
    | patch_{x,y}                    | id of the patch within the tract      |
    +--------------------------------+---------------------------------------+
    | cell_{x,y}                     | id of the cell withing the patch      |
    +--------------------------------+---------------------------------------+
    | SNR                            | Signal-to-Noise of source             |
    +--------------------------------+---------------------------------------+
    | {catType}_band_flux_{band}     | Flux measuremnt in the reference band |
    +--------------------------------+---------------------------------------+
    | {catType}_band_flux_err_{band} | Flux error in the reference band      |
    +--------------------------------+---------------------------------------+
    | {catType}_g_{i}                | Shear measurements                    |
    +--------------------------------+---------------------------------------+


    Two additional tables are produced beyond the tables produced by
    the base :py:class:`hpmcm.Match` class

    +----------------+---------------------------------------------------+
    | Key            | Class                                             |
    +================+===================================================+
    | _object_shear  | :py:class:`hpmcm.output_tables.ShearTable`        |
    +----------------+---------------------------------------------------+
    | _cluster_shear | :py:class:`hpmcm.output_tables.ShearTable`        |
    +----------------+---------------------------------------------------+
    """

    inputTableClass: type = input_tables.ShearCoaddSourceTable
    extraCols: list[str] = ["ra", "dec", "xPix", "yPix", "g_1", "g_2"]

    def __init__(
        self,
        **kwargs: Any,
    ):
        self.pixelMatchScale: int = kwargs.get("pixelMatchScale", 1)
        self.catType: str = kwargs.get("catalogType", "wmom")
        self.deshear: float | None = kwargs.get("deshear", None)
        Match.__init__(self, **kwargs)

    @classmethod
    def createShearMatch(
        cls,
        **kwargs: Any,
    ) -> ShearMatch:
        """Helper function to create a `ShearMatch` object

        This will use the use pixel-coordinates read from
        the input shear tables.

        Parameters
        ----------
        kwargs:
            Passed directly to `ShearMatch` constructor.

        Returns
        -------
        Object to create matches for the requested region
        """
        nPix = np.array([30000, 30000])
        kw = dict(
            pixelSize=0.2 / 3600.0,
            nPixels=nPix,
            cellSize=150,
            cellBuffer=25,
            cellMaxObject=1000,
        )
        return cls(**kw, **kwargs)

    def getCellIndices(
        self,
        df: pandas.DataFrame,
    ) -> np.ndarray:
        """Get the cell index assocatiated to each source"""
        return (self.nCell[1] * df["cellIdxX"] + df["cellIdxY"]).astype(int)

    def _buildCellData(
        self,
        idOffset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
    ) -> CellData:
        return ShearCellData(self, idOffset, corner, size, idx, self.cellBuffer)

    def extractShearStats(self) -> list[pandas.DataFrame]:
        """Extract shear stats

        Theis will produce two :py:class:`hpmcm.output_tables.ShearTable`,
        one for the objects, and the other for the clusters.
        """
        clusterShearStatsTables = []
        objectShearStatsTables = []

        for ix in range(int(self.nCell[0])):
            for iy in range(int(self.nCell[1])):
                iCell = self.getCellIdx(ix, iy)
                if iCell not in self.cellDict:
                    continue
                cellData = self.cellDict[iCell]
                assert isinstance(cellData, ShearCellData)
                clusterShearStatsTables.append(
                    output_tables.ShearTable.buildClusterShearStats(cellData).data
                )
                objectShearStatsTables.append(
                    output_tables.ShearTable.buildObjectShearStats(cellData).data
                )

        return [
            pandas.concat(clusterShearStatsTables),
            pandas.concat(objectShearStatsTables),
        ]

    def _getPixValues(self, df: pandas.DataFrame) -> tuple[np.ndarray, np.ndarray]:
        xPix, yPix = (
            df["xPix"].values,
            df["yPix"].values,
        )
        return xPix, yPix

    def reduceDataFrame(
        self,
        df: pandas.DataFrame,
    ) -> pandas.DataFrame:
        """Reduce a single input DataFrame

        Notes
        -----

        This applies a trivial cut on signal-to-noise (SNR>1).

        This will add these columns to the output dataframes

        +--------------+-------------------------------------+
        | Column       | Description                         |
        +==============+=====================================+
        | id           | Index of object inside catalog      |
        +--------------+-------------------------------------+
        | ra           | Source RA                           |
        +--------------+-------------------------------------+
        | dec          | Source DEC                          |
        +--------------+-------------------------------------+
        | cellIdxX     | X-index of Cell                     |
        +--------------+-------------------------------------+
        | cellIdxY     | Y-index of Cell                     |
        +--------------+-------------------------------------+
        | xCellCoadd   | X-coordinate in cell frame          |
        +--------------+-------------------------------------+
        | yCellCoadd   | Y-coordinate in cell frame          |
        +--------------+-------------------------------------+
        | xPix         | X-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | yPix         | Y-coordinate in global WCS frame    |
        +--------------+-------------------------------------+
        | g_1          | Shear g_1 component estimate        |
        +--------------+-------------------------------------+
        | g_2          | Shear g_2 component estimate        |
        +--------------+-------------------------------------+
        | SNR          | Signal-to-noise ratio               |
        +--------------+-------------------------------------+

        """
        df_clean = df[(df.SNR > 1)]
        df_red = df_clean.copy(deep=True)

        return df_red[
            [
                "id",
                "ra",
                "dec",
                "xPix",
                "yPix",
                "xCellCoadd",
                "yCellCoadd",
                "SNR",
                "g_1",
                "g_2",
                "cellIdxX",
                "cellIdxY",
            ]
        ]



"""
The :py:mod:`hpmcm` package collects a set of tools to do catalog
cross matching using the following strategy.

1. The match region is mapped onto a single WCS and each sources
   is projected into pixel coordinates in that WCS.  This is defined
   using a sub-class of :py:class:`hpmcm.match.Match`

2. The match region is sub-divided into
   :py:class:`cells <hpmcm.cell.CellData>`, which include some
   overlap region between them.

3. For each cell, a "counts map" of sources in that cell is produced,
   counting the number of sources in each pixel of the projected WCS.

4. A clustering algorithm is run to extract the "footprints" of sets
   of adjacent pixels that contain sources.   Those are then used to
   create a set of :py:class:`clusters <hpmcm.cluster.ClusterData>`

5. Each cluster in then refined into
   :py:class:`objects <hpmcm.cluster.ObjectData>`, by repeating the
   counts map and footprint detection using progressive smaller pixels
   until each objects consists only of sources within the requested
   match radius.  At each iteration the pixel scale is halved.

6. Information about the source-to-object and source-to-clusters
   associations, as well information about the resulting objects
   and clusters are collected into :py:mod:`hpmcm.output_tables`


The two sub-classes of `Match` are:

1. :py:class:`WcsMatch <hpmcm.wcs_match.WcsMatch>` allows the user to
   define the WCS used for the projection and takes any number of input
   catalogs.

2. :py:class:`ShearMatch <hpmcm.shear_match.ShearMatch>` uses a predefined
   WCS that was used to create the meta-detection ShearObject catalogs, and
   expects exactly 5 input catalogs: a reference catalog and 4 counterfactual
   shear catalogs.
"""

from . import classify, match_utils, package_utils, shear_utils, utils, viz_utils
from ._version import __version__
from .cell import CellData, ShearCellData
from .cluster import ClusterData, ShearClusterData
from .footprint import Footprint, FootprintSet
from .match import Match
from .object import ObjectData, ShearObjectData
from .output_tables import (
    ClusterAssocTable,
    ClusterStatsTable,
    ObjectAssocTable,
    ObjectStatsTable,
    ShearTable,
)
from .shear_data import ShearData
from .shear_match import ShearMatch
from .table import TableColumnInfo, TableInterface
from .wcs_match import WcsMatch

__all__ = [
    "CellData",
    "ClusterAssocTable",
    "ClusterData",
    "ClusterStatsTable",
    "Footprint",
    "FootprintSet",
    "Match",
    "ObjectAssocTable",
    "ObjectData",
    "ObjectStatsTable",
    "ShearCellData",
    "ShearClusterData",
    "ShearData",
    "ShearTable",
    "ShearObjectData",
    "ShearMatch",
    "TableColumnInfo",
    "TableInterface",
    "WcsMatch",
    "classify",
    "match_utils",
    "package_utils",
    "shear_utils",
    "utils",
    "viz_utils",
    "__version__",
]



import pytest

from hpmcm import package_utils


@pytest.fixture(name="setup_data", scope="package")
def setup_data(request: pytest.FixtureRequest) -> int:
    """Download test data and unpack it"""
    ret_val = package_utils.setupTestDataArea()

    request.addfinalizer(package_utils.teardownTestDataArea)

    return ret_val



from __future__ import annotations


import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure, SubFigure

from .cell import CellData
from .cluster import ClusterData
from .match import Match


def showShearObjs(matcher: Match, iK: tuple[int, int]) -> Figure | SubFigure:
    """Draw the objects in a cluster

    Parameters
    ----------
    matcher:
        Match object

    iK:
        Indices of the particular cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    cellData = matcher.cellDict[iK[0]]
    cluster = cellData.clusterDict[iK[1]]
    extent = cluster.footprint.extent()
    cluster.extract(cellData)
    xOffset = cellData.minPix[0] + 25
    yOffset = cellData.minPix[1] + 25
    assert cluster.data is not None
    xOff = cluster.data.xPix - xOffset
    yOff = cluster.data.yPix - yOffset
    catIndices = cluster.catIndices
    image = cluster.footprint.cutout
    img = plt.imshow(image, origin="lower", extent=extent)
    colors = ["red", "blue", "green", "cyan", "orange"]
    markers = [".", "<", ">", "v", "^"]
    for iObj, obj in enumerate(cluster.objects):
        for x_, y_, i_ in zip(xOff[obj.mask], yOff[obj.mask], catIndices[obj.mask]):
            img.axes.scatter(x_, y_, c=colors[iObj % 5], marker=markers[i_ % 5])
    _cb = plt.colorbar(label="Objects per pixel")
    assert img.axes.figure is not None
    return img.axes.figure


def showShearObj(matcher: Match, iK: tuple[int, int]) -> Figure | SubFigure:
    """Draw a single object

    Parameters
    ----------
    matcher:
        Match object

    iK:
        Indices of the particular object

    Returns
    -------
    Figure showing the object in question
    """

    cellData = matcher.cellDict[iK[0]]
    theObj = cellData.objectDict[iK[1]]
    cluster = theObj.parentCluster
    extent = cluster.footprint.extent()
    cluster.extract(cellData)
    xOffset = cellData.minPix[0] + 25
    yOffset = cellData.minPix[1] + 25
    assert cluster.data is not None
    xOff = cluster.data.xPix - xOffset
    yOff = cluster.data.yPix - yOffset
    catIndices = cluster.catIndices
    img = plt.imshow(cluster.footprint.cutout, origin="lower", extent=extent)
    markers = [".", "<", ">", "v", "^"]
    for _iObj, obj in enumerate(cluster.objects):
        if obj.objectId == theObj.objectId:
            color = "red"
        else:  # pragma: no cover
            color = "blue"
        for x_, y_, i_ in zip(xOff[obj.mask], yOff[obj.mask], catIndices[obj.mask]):
            img.axes.scatter(x_, y_, c=color, marker=markers[i_ % 5])
    _cb = plt.colorbar(label="Objects per pixel")
    assert img.axes.figure is not None
    return img.axes.figure


def showCluster(
    image: np.ndarray,
    cluster: ClusterData,
    cellData: CellData,
    mask: np.ndarray | None = None,
) -> Figure | SubFigure:
    """Draw a cluster

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cellData:
        Parent Cell for the cluster

    mask:
        Mask showing which sources are in the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cellData)
    xOffset = cellData.minPix[0] + 25
    yOffset = cellData.minPix[1] + 25
    xOff = cluster.xCluster
    yOff = cluster.yCluster
    if mask is not None:  # pragma: no cover
        xOff = xOff[mask]
        yOff = yOff[mask]
    xC = cluster.xCent - xOffset
    yC = cluster.yCent - yOffset

    img = plt.imshow(
        image[cluster.footprint.sliceX][cluster.footprint.sliceY],
        origin="lower",
        extent=extent,
        cmap="grey",
    )
    _cb = plt.colorbar(label="Objects per pixel")
    try:
        assert cluster.data is not None
        xOff_u = xOff - cluster.data.dxShear
        yOff_u = yOff - cluster.data.dyShear
        img.axes.scatter(xOff_u, yOff_u, marker="x")
    except Exception:  # pragma: no cover
        pass
    img.axes.scatter(xOff, yOff)
    img.axes.scatter(xC, yC, marker="+", c="green")
    img.axes.set_xlabel("x [pixels]")
    img.axes.set_ylabel("y [pixels]")
    assert img.axes.figure is not None
    return img.axes.figure


def showObjects(
    image: np.ndarray,
    cluster: ClusterData,
    cellData: CellData,
) -> Figure | SubFigure:
    """Draw a cluster, showing the objects

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cellData:
        Parent Cell for the cluster

    mask:
        Mask showing which sources are in the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cellData)
    xOffset = cellData.minPix[0]
    yOffset = cellData.minPix[1]
    xOff = cluster.xCluster
    yOff = cluster.yCluster
    img = plt.imshow(
        image[cluster.footprint.sliceX][cluster.footprint.sliceY],
        origin="lower",
        extent=extent,
    )
    _cb = plt.colorbar(label="Objects per pixel")
    colors = ["red", "blue", "green", "cyan", "orange"]
    for iObj, obj in enumerate(cluster.objects):
        xC = obj.xCent - xOffset
        yC = obj.yCent - yOffset
        img.axes.scatter(
            xOff[obj.mask],
            yOff[obj.mask],
            c=colors[iObj % 5],
            s=1 + np.ceil(iObj / 5),
        )
        print(1 + np.ceil(iObj / 5))
        img.axes.scatter(xC, yC, marker="+", c=colors[iObj % 6])
    assert img.axes.figure is not None
    return img.axes.figure


def showObjectsV2(
    image: np.ndarray,
    cluster: ClusterData,
    cellData: CellData,
) -> Figure | SubFigure:
    """Draw a cluster, showing the objects

    Parameters
    ----------
    image:
        Counts map used to make clusters

    cluster:
        Cluster being draw

    cellData:
        Parent Cell for the cluster

    Returns
    -------
    Figure showing the cluster in question
    """
    extent = cluster.footprint.extent()
    cluster.extract(cellData)
    xOffset = cellData.minPix[0]
    yOffset = cellData.minPix[1]
    xOff = cluster.xPix - xOffset
    yOff = cluster.yPix - yOffset
    img = plt.imshow(
        image[cluster.footprint.sliceX][cluster.footprint.sliceY],
        origin="lower",
        extent=extent,
    )
    _cb = plt.colorbar()
    colors = ["red", "blue", "green", "cyan", "orange"]
    for xOff_, yOff_, iCat_ in zip(xOff, yOff, cluster.catIndices):
        if iCat_ % 5 == 0 and iCat_ != 20:
            continue
        img.axes.scatter(
            xOff_, yOff_, c=colors[iCat_ % 5], s=20 - 3 * np.ceil(iCat_ / 5)
        )
    assert img.axes.figure is not None
    return img.axes.figure



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import shear_utils
from .footprint import Footprint
from .object import ObjectData

if TYPE_CHECKING:
    from .cell import CellData


class ClusterData:
    """Class that defines clusters

    A cluster is a set of sources within a footprint of
    adjacent pixels.

    Attributes
    ----------
    iCluster : int
        Cluster ID

    footprint: Footprint
        Footprint of this cluster in the CellData counts map

    origCluster : int
        Id of the original cluster this cluster was made from

    sources: np.ndarray
        Data about the sources in this cluster

    nSrc : int
        Number of sources in this cluster

    nUnique : int
        Number of catalogs contributing sources to this cluster

    objects: list[ObjectData]
        Data about the objects in this cluster

    xCent : float
        X-pixel value of cluster centroid (in WCS used to do matching)

    yCent : float
        Y-pixel value of cluster centroid (in WCS used to do matching)

    rmsDist: float
        RMS distance of sources to the centroid

    snrMean: float
        mean signal-to-noise

    snrRms: float
        RMS of signal-to-noise

    dist2: np.ndarray
        Distances of sources to the centroid
    """

    def __init__(
        self,
        iCluster: int,
        footprint: Footprint,
        sources: np.ndarray,
        origCluster: int | None = None,
    ):
        """Build from a Footprint and data about the
        sources in that Footprint"""
        self.iCluster: int = iCluster
        self.footprint: Footprint = footprint
        self.origCluster: int = iCluster
        if origCluster is not None:
            self.origCluster = origCluster
        self.sources: np.ndarray = sources

        self.nSrc: int = self.sources[0].size
        self.nUnique: int = len(np.unique(self.sources[0]))
        self.objects: list[ObjectData] = []
        self.data: pandas.DataFrame | None = None

        self.xCent: float = np.nan
        self.yCent: float = np.nan
        self.rmsDist: float = np.nan
        self.snrMean: float = np.nan
        self.snrRms: float = np.nan
        self.dist2: np.ndarray = np.array([])
        self.pixelMatchScale: int = 1

    def extract(self, cellData: CellData) -> None:
        """Extract the xPix, yPix and snr data from
        the sources in this cluster
        """
        xOffset = self.footprint.sliceX.start * self.pixelMatchScale
        yOffset = self.footprint.sliceY.start * self.pixelMatchScale

        seriesList = []

        for _i, (iCat, srcIdx) in enumerate(zip(self.sources[0], self.sources[2])):
            seriesList.append(cellData.data[iCat].iloc[srcIdx])

        self.data = pandas.DataFrame(seriesList)
        self.data["iCat"] = self.sources[0]
        self.data["srcId"] = self.sources[1]
        self.data["srcIdx"] = self.sources[2]
        self.data["xCluster"] = self.data.xCell - xOffset
        self.data["yCluster"] = self.data.yCell - yOffset

    @property
    def catIndices(self) -> np.ndarray:
        """Return the catalog indices for all the sources in the cluster"""
        assert self.data is not None
        return self.data.iCat

    @property
    def srcIds(self) -> np.ndarray:
        """Return the ids for all the sources in the cluster"""
        assert self.data is not None
        return self.data.srcId

    @property
    def srcIdxs(self) -> np.ndarray:
        """Return the indices for all the sources in the cluster"""
        assert self.data is not None
        return self.data.srcIdx

    def hasRefCatalog(self, refCatId: int = 0) -> bool:
        """Is there a source from the reference catalog"""
        return refCatId in self.sources[0]

    @property
    def xCluster(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the footprint"""
        assert self.data is not None
        return self.data.xCluster

    @property
    def yCluster(self) -> np.ndarray:
        """Return the y-positions of the soures w.r.t. the footprint"""
        assert self.data is not None
        return self.data.yCluster

    @property
    def xPix(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the WCS"""
        assert self.data is not None
        return self.data.xPix

    @property
    def yPix(self) -> np.ndarray:
        """Return the x-positions of the soures w.r.t. the WCS"""
        assert self.data is not None
        return self.data.yPix

    def addObject(
        self, cellData: CellData, mask: np.ndarray | None = None
    ) -> ObjectData:
        """Add a new object to this cluster"""
        newObject = cellData.addObject(self, mask)
        self.objects.append(newObject)
        return newObject


class ShearClusterData(ClusterData):
    """Subclass of ClusterData that can compute shear statisitics

    Attributes
    ----------
    pixelMatchScale: int
        Number of pixel merged in the original counts map
    """

    def __init__(
        self,
        iCluster: int,
        footprint: Footprint,
        sources: np.ndarray,
        origCluster: int | None = None,
        pixelMatchScale: int = 1,
    ):
        ClusterData.__init__(self, iCluster, footprint, sources, origCluster)
        self.pixelMatchScale = pixelMatchScale

    def shearStats(self) -> dict:
        """Return the shear statistics"""
        assert self.data is not None
        return shear_utils.shearStats(self.data)



from __future__ import annotations


import numpy as np
import pandas

from .footprint import FootprintSet


def findClusterIdsFromArrays(
    xLocals: np.ndarray,
    yLocals: np.ndarray,
    clusterKey: np.ndarray,
) -> np.ndarray:
    """Associate sources to clusters using `clusterkey`
    which is a map where any pixel associated to a cluster
    has the cluster index as its value

    Parameters
    ----------
    xLocals:
        Local pixel x-positions

    yLocals:
        Local pixel y-positions

    clusterKey:
        2D-map of cluster Ids by pixel position

    Returns
    -------
    Ids of associated clusters
    """
    return np.array(
        [clusterKey[xLocal, yLocal] for xLocal, yLocal in zip(xLocals, yLocals)]
    ).astype(np.int32)


def findClusterIds(
    df: pandas.DataFrame,
    clusterKey: np.ndarray,
    pixelMatchScale: int = 1,
) -> np.ndarray:
    """Associate sources to clusters using `clusterkey`
    which is a map where any pixel associated to a cluster
    has the cluster index as its value

    Parameters
    ----------
    df:
        DataFrame with local pixel positions (xCell, yCell)

    clusterKey:
        2D-map of cluster Ids by pixel position

    pixelMatchScale:
        Scale-factor to use in making cluster map

    Returns
    -------
    Ids of associated clusters
    """
    return findClusterIdsFromArrays(
        np.floor(df["xCell"] / pixelMatchScale).astype(int),
        np.floor(df["yCell"] / pixelMatchScale).astype(int),
        clusterKey,
    )


def fillCountsMapFromArrays(
    xLocals: np.ndarray,
    yLocals: np.ndarray,
    nPix: np.ndarray,
    weights: np.ndarray | None = None,
) -> np.ndarray:
    """Fill a source counts map

    Parameters
    ----------
    xLocals:
        Local pixel x-positions

    yLocals:
        Local pixel y-positions

    nPix:
        Number of pixels in x,y for counts map

    weights:
        If provided, weights to apply for each entry in counts map

    Returns
    -------
    Counts map of source in cell, projected into nPix,nPix grid
    """
    hist = np.histogram2d(
        xLocals,
        yLocals,
        bins=(nPix[0], nPix[1]),
        range=((0, nPix[0]), (0, nPix[1])),
        weights=weights,
    )
    return hist[0]


def fillCountsMapFromDf(
    df: pandas.DataFrame,
    nPix: np.ndarray,
    weightName: str | None = None,
    pixelMatchScale: int = 1,
) -> np.ndarray:
    """Fill a source counts map from a reduced dataframe for one input
    catalog

    Parameters
    ----------
    df:
        DataFrame with local pixel positions (xCell, yCell)

    nPix:
        Number of pixels in x,y for counts map

    weightName:
        If provided column to use for weights

    pixelMatchScale:
        Scale-factor to use in making cluster map

    Returns
    -------
    Counts map of source in cell, projected into nPix,nPix grid
    """
    if weightName is None:
        weights = None
    else:  # pragma: no cover
        weights = df[weightName].values
    return fillCountsMapFromArrays(
        df["xCell"] / pixelMatchScale,
        df["yCell"] / pixelMatchScale,
        nPix=np.ceil(nPix / pixelMatchScale).astype(int),
        weights=weights,
    )


def getFootprints(
    countsMap: np.ndarray,
    buf: int,
    pixelMatchScale: int = 1,
) -> dict:
    """Take a source counts map and do clustering using Footprint detection

    Parameters
    ----------
    countsMap:
        Map of source counts

    buf:
        Number of pixels in cell-edge buffer

    pixelMatchScale:
        Scale-factor used in making cluster map

    Returns
    -------
    Footprint data


    Notes
    -----
    image: np.ndarray : countsMap

    footprints: FootprintSet : Clustering FootprintSet

    footprintKey: np.ndarray : Array with same shape as countsMap, with cluster associations
    """
    footprintsOrig = FootprintSet.detect(countsMap)
    footprints = footprintsOrig.filter(buf, pixelMatchScale)
    footprintKey = footprints.fpKey
    return dict(image=countsMap, footprints=footprints, footprintKey=footprintKey)


def associateSourcesToFootprints(
    data: list[pandas.DataFrame],
    clusterKey: np.ndarray,
    pixelMatchScale: int = 1,
) -> list[np.ndarray]:
    """Loop through data and associate sources to clusters

    Parameters
    ----------
    data:
        Input DataFrames

    clusterKey:
        2D-map of cluster Ids by pixel position

    pixelMatchScale:
        Scale-factor used in making cluster map

    Returns
    -------
    Lists of clusters associated to each source
    output[i][j] will give the id of the cluster associated
    to source j in input catalog i.
    """
    return [findClusterIds(df, clusterKey, pixelMatchScale) for df in data]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from . import utils
from .cluster import ClusterData
from .object import ObjectData

if TYPE_CHECKING:
    from .cell import CellData


def heirarchicalProcessObject(
    objData: ObjectData, cellData: CellData, pixelR2Cut: float, recurse: int = 0
) -> None:
    """Recursively process an object and make sub-objects

    Parameters
    ----------
    objData:
        Object being processed

    cellData:
        Associated cell

    pixelR2Cut:
        Distance cut for associated, in pixels**2

    recurse:
        Current recursion level


    Notes
    -----
    This function will test if the input objData is good,
    i.e., that all the sources lie within the pixelR2Cut,
    and will split objData into multiple objects if
    some of the sources lie outside the cut

    The new objects will be added to the parent cluster of
    the original object
    """
    if recurse > cellData.matcher.maxSubDivision:
        return
    objData.recurse = recurse

    if objData.nSrc == 0:
        print("Empty object", objData.nSrc, objData.nUnique, recurse)
        return

    assert objData.data is not None
    if objData.mask.sum() == 1:
        objData.xCent = objData.data.xPix.values[0]
        objData.yCent = objData.data.yPix.values[0]
        objData.dist2 = np.zeros((1), float)
        objData.rmsDist = 0.0
        objData.snrMean = objData.data.SNR.values[0]
        objData.snrRms = 0.0
        return

    sumSnr = np.sum(objData.data.SNR)
    objData.xCent = np.sum(objData.data.xPix * objData.data.SNR) / sumSnr
    objData.yCent = np.sum(objData.data.yPix * objData.data.SNR) / sumSnr
    objData.dist2 = np.array(
        (objData.xCent - objData.data.xPix) ** 2
        + (objData.yCent - objData.data.yPix) ** 2
    )
    objData.rmsDist = np.sqrt(np.mean(objData.dist2))
    objData.snrMean = np.mean(objData.data.SNR.values)
    objData.snrRms = np.std(objData.data.SNR.values)

    subMask = objData.dist2 < pixelR2Cut
    if subMask.all():
        return

    if recurse >= cellData.matcher.maxSubDivision:
        return

    heirarchicalSplitObject(objData, cellData, pixelR2Cut, recurse=recurse + 1)
    return


def heirarchicalSplitObject(
    objData: ObjectData, cellData: CellData, pixelR2Cut: float, recurse: int = 0
) -> None:
    """Split up a cluster keeping only one source per input
    catalog

    Parameters
    ----------
    objData:
        Object being processed

    cellData:
        Associated cell

    pixelR2Cut:
        Distance cut for associated, in pixels**2

    recurse:
        Current recursion level

    Notes
    -----
    This function actually does the splitting of objects
    """

    if recurse > cellData.matcher.maxSubDivision:
        return
    objData.recurse = recurse
    objData.extract()

    assert objData.data is not None

    sliceX = objData.parentCluster.footprint.sliceX
    sliceY = objData.parentCluster.footprint.sliceY
    zoomFactors = [1, 2, 4, 8, 16, 32]
    zoomFactor = zoomFactors[recurse]

    nPix = zoomFactor * np.array(
        [sliceX.stop - sliceX.start, sliceY.stop - sliceY.start]
    )
    zoomX = zoomFactor * objData.data.xCluster / objData.parentCluster.pixelMatchScale
    zoomY = zoomFactor * objData.data.yCluster / objData.parentCluster.pixelMatchScale

    countsMap = utils.fillCountsMapFromArrays(zoomX, zoomY, nPix)

    fpDict = utils.getFootprints(countsMap, buf=0)
    footprints = fpDict["footprints"]
    nFootprints = len(footprints.footprints)
    if nFootprints == 1:
        if recurse >= cellData.matcher.maxSubDivision:
            return
        heirarchicalSplitObject(objData, cellData, pixelR2Cut, recurse=recurse + 1)
        return

    footprintKey = fpDict["footprintKey"]
    footprintIds = utils.findClusterIdsFromArrays(
        zoomX.values.astype(int), zoomY.values.astype(int), footprintKey
    )

    biggest = np.argmax(np.bincount(footprintIds))
    biggestMask = np.zeros(objData.mask.shape, dtype=bool)

    for iFp in range(nFootprints):
        subMask = footprintIds == iFp
        count = 0
        newMask = np.zeros(objData.mask.shape, dtype=bool)
        for i, val in enumerate(objData.mask):
            if val:
                newMask[i] = subMask[count]
                count += 1
        assert subMask.sum() == newMask.sum()

        if iFp == biggest:
            biggestMask = newMask
            continue

        newObject = objData.parentCluster.addObject(cellData, newMask)
        heirarchicalProcessObject(newObject, cellData, pixelR2Cut, recurse=recurse)

    objData.mask = biggestMask
    objData.extract()
    heirarchicalProcessObject(objData, cellData, pixelR2Cut, recurse=recurse)
    return


def heirarchicalProcessCluster(
    cluster: ClusterData, cellData: CellData, pixelR2Cut: float
) -> list[ObjectData]:
    """Function that is called recursively to
    split clusters until they consist only of sources within
    the match radius of the cluster centroid.

    Recursively process a cluster and make associated objects

    Parameters
    ----------
    cluster:
        Cluster being processed

    cellData:
        Associated cell

    pixelR2Cut:
        Distance cut for associated, in pixels**2

    Returns
    -------
    Objects produced in cluster processing
    """
    cluster.extract(cellData)
    assert cluster.data is not None

    cluster.nSrc = len(cluster.data.iCat)
    cluster.nUnique = len(np.unique(cluster.data.iCat.values))
    if cluster.nSrc == 0:
        print("Empty cluster", cluster.nSrc, cluster.nUnique)
        return cluster.objects

    if cluster.nSrc == 1:
        cluster.xCent = cluster.data.xPix.values[0]
        cluster.yCent = cluster.data.yPix.values[0]
        cluster.dist2 = np.zeros((1))
        cluster.rmsDist = 0.0
        cluster.snrMean = cluster.data.SNR.values[0]
        cluster.snrRms = 0.0
        initialObject = cluster.addObject(cellData)
        heirarchicalProcessObject(initialObject, cellData, pixelR2Cut)
        return cluster.objects

    sumSnr = np.sum(cluster.data.SNR)
    cluster.xCent = np.sum(cluster.data.xPix * cluster.data.SNR) / sumSnr
    cluster.yCent = np.sum(cluster.data.yPix * cluster.data.SNR) / sumSnr
    cluster.dist2 = (cluster.xCent - cluster.data.xPix) ** 2 + (
        cluster.yCent - cluster.data.yPix
    ) ** 2
    cluster.rmsDist = np.sqrt(np.mean(cluster.dist2))
    cluster.snrMean = np.mean(cluster.data.SNR.values)
    cluster.snrRms = np.std(cluster.data.SNR.values)

    initialObject = cluster.addObject(cellData)
    heirarchicalProcessObject(initialObject, cellData, pixelR2Cut)
    return cluster.objects



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import pandas
import tables_io

from .shear_data import ShearData

if TYPE_CHECKING:
    from .cell import CellData
    from .shear_match import ShearMatch


# These are the names of the various shear catalogs
SHEAR_NAMES = ["ns", "2p", "2m", "1p", "1m"]

# These are the coeffs for the various shear catalogs
DESHEAR_COEFFS = np.array(
    [
        [0, 0, 0, 0],
        [0, 1, 1, 0],
        [0, -1, -1, 0],
        [1, 0, 0, -1],
        [-1, 0, 0, 1],
    ]
)


def shearStats(df: pandas.DataFrame) -> dict:
    """Return the shear statistics

    {st} is the shear type, one of "gauss", "pgauss", "wmom"

    {i}, {j} index the shear parameters 1, 2

    Parameters
    ----------
    df:
        Input DataFrame, must have :py:class:`hpmcm.ShearTable` schema

    Returns
    -------
    Shear stats in a dict.

    Notes
    -----
    Shear stats include:

    +-----------------+-----------------------------------------------------+
    | Key             | Description                                         |
    +=================+=====================================================+
    | n_{st}          | Number of sources from that catalog                 |
    +-----------------+-----------------------------------------------------+
    | g_{i}_{st}      | g_{i} shear parameter for that catalog              |
    +-----------------+-----------------------------------------------------+
    | delta_g_{i}_{j} | g_{i,j} shear measurment: g_{i}_{j}p - g_{i}_{j}m   |
    +-----------------+-----------------------------------------------------+
    | good            | True if every catalog has one source in this object |
    +-----------------+-----------------------------------------------------+

    If the matching is not good, then delta_g_1 = delta_g_2 = np.nan
    """
    outDict: dict[str, float | int] = {}
    allGood = True
    for i, name_ in enumerate(SHEAR_NAMES):
        mask = df.iCat == i
        nCat = mask.sum()
        if nCat != 1:
            allGood = False
        outDict[f"n_{name_}"] = int(nCat)
        if nCat:
            outDict[f"g_1_{name_}"] = df[mask].g_1.values.mean()
            outDict[f"g_2_{name_}"] = df[mask].g_2.values.mean()
        else:
            outDict[f"g_1_{name_}"] = np.nan
            outDict[f"g_2_{name_}"] = np.nan
    if allGood:
        outDict["delta_g_1_1"] = outDict["g_1_1p"] - outDict["g_1_1m"]
        outDict["delta_g_2_2"] = outDict["g_2_2p"] - outDict["g_2_2m"]
        outDict["delta_g_1_2"] = outDict["g_1_2p"] - outDict["g_1_2m"]
        outDict["delta_g_2_1"] = outDict["g_2_1p"] - outDict["g_2_1m"]
    else:
        outDict["delta_g_1_1"] = np.nan
        outDict["delta_g_2_2"] = np.nan
        outDict["delta_g_1_2"] = np.nan
        outDict["delta_g_2_1"] = np.nan
    outDict["good"] = allGood
    return outDict


def shearReport(
    basefile: str,
    outputFileBase: str | None,
    shear: float,
    catType: str,
    tract: int,
    snrCut: float = 7.5,
) -> None:
    """Report on the shear calibration

    Parameters
    ----------
    basefile
        Input base file name (see notes)

    outputFileBase:
        Output file name (see notes)

    shear:
        Applied shear

    catType:
        Catalog type (one of ["pgauss", "gauss", "wmom"]

    tract:
        Tract, written to outout data

    snrCut:
        Signal-to-noise cut.

    Notes
    -----
    This will read the object shear data from "{basefile}_cluster_shear.pq"
    This will read the object statisticis from "{basefile}_cluster_stats.pq"

    This will write the shear stats to "{outputFileBase}.pkl"
    This will write the figures to "{outputFileBase}_{figure}.png"
    """
    t = tables_io.read(f"{basefile}_cluster_shear.pq")
    t2 = tables_io.read(f"{basefile}_cluster_stats.pq")

    shearData = ShearData(t, t2, shear, catType, tract, snrCut=snrCut)

    if outputFileBase is not None:
        shearData.save(f"{outputFileBase}.pkl")
        shearData.savefigs(outputFileBase)


def mergeShearReports(
    inputs: list[str],
    outputFile: str,
) -> None:
    """Merge reports on the shear calibration

    Parameters
    ----------
    inputs:
        List of input ShearData pickle files

    outputFile:
        Where to write the merged file
    """
    outDict: dict[str, Any] = {}
    for input_ in inputs:
        shearData = ShearData.load(input_)
        inputDict = shearData.toDict()
        for key, val in inputDict.items():
            if key in outDict:
                outDict[key].append(val)
            else:
                outDict[key] = [val]

    outDF = pandas.DataFrame(outDict)
    outDF.to_parquet(outputFile)


def splitByTypeAndClean(
    basefile: str,
    tract: int,
    shear: float,
    catType: str,
    *,
    clean: bool = False,
) -> None:  # pragma: no cover
    """Split a parquet file by shear catalog type and tract

    Parameters
    ----------
    basefile:
        Original file name

    tract:
        Tract to select

    shear:
        Applied shear, saved to output

    catType:
        Catalog type to select

    clean:
        Remove duplicates

    Notes
    -----
    This will create 5 files with the pattern:
    "{basefile}_uncleaned_{tract}_{type}.pq"

    +--------------+-------------------------------------+
    | Column       | Description                         |
    +==============+=====================================+
    | id           | Index of object inside catalog      |
    +--------------+-------------------------------------+
    | orig_id      | Original object id                  |
    +--------------+-------------------------------------+
    | cellIdxX     | X-index of Cell                     |
    +--------------+-------------------------------------+
    | cellIdxY     | Y-index of Cell                     |
    +--------------+-------------------------------------+
    | xCellCoadd   | X-coordinate in cell frame          |
    +--------------+-------------------------------------+
    | yCellCoadd   | Y-coordinate in cell frame          |
    +--------------+-------------------------------------+
    | xPix         | X-coordinate in global WCS frame    |
    +--------------+-------------------------------------+
    | yPix         | Y-coordinate in global WCS frame    |
    +--------------+-------------------------------------+
    | g_1          | Shear g_1 component estimate        |
    +--------------+-------------------------------------+
    | g_2          | Shear g_2 component estimate        |
    +--------------+-------------------------------------+
    | SNR          | Signal-to-noise ratio               |
    +--------------+-------------------------------------+

    """
    p = tables_io.read(basefile)
    if clean:
        clean_st = "cleaned"
        cellCut = 75
    else:
        clean_st = "uncleaned"
        cellCut = 80
    for type_ in SHEAR_NAMES:
        mask = p["shear_type"] == type_
        sub = p[mask].copy(deep=True)
        cellIdxX = (20 * sub["patch_x"].values + sub["cell_x"].values).astype(int)
        cellIdxY = (20 * sub["patch_y"].values + sub["cell_y"].values).astype(int)
        cent_x = 150 * cellIdxX - 75
        cent_y = 150 * cellIdxY - 75
        xCellCoadd = sub["col"] - cent_x
        yCellCoadd = sub["row"] - cent_y
        sub["xPix"] = sub["col"] + 25
        sub["yPix"] = sub["row"] + 25
        sub["xCellCoadd"] = xCellCoadd
        sub["yCellCoadd"] = yCellCoadd
        sub["SNR"] = sub[f"{catType}_band_flux_r"] / sub[f"{catType}_band_flux_err_r"]
        sub["g_1"] = sub[f"{catType}_g_1"]
        sub["g_2"] = sub[f"{catType}_g_2"]
        sub["cellIdxX"] = cellIdxX
        sub["cellIdxY"] = cellIdxY
        sub["orig_id"] = sub.id
        sub["id"] = np.arange(len(sub))
        central_to_cell = np.bitwise_and(
            np.fabs(xCellCoadd) < cellCut, np.fabs(yCellCoadd) < cellCut
        )
        central_to_patch = np.bitwise_and(
            np.fabs(sub["cell_x"].values - 10.5) < 10,
            np.fabs(sub["cell_y"].values - 10.5) < 10,
        )
        right_tract = sub["tract"] == tract
        central = np.bitwise_and(central_to_cell, central_to_patch)
        selected = np.bitwise_and(right_tract, central)
        cleaned = sub[selected].copy(deep=True)
        cleaned["shear"] = np.repeat(shear, len(cleaned))
        cleaned.to_parquet(basefile.replace(".parq", f"_{clean_st}_{tract}_{type_}.pq"))


def reduceShearDataForCell(
    cell: CellData, iCat: int, dataframe: pandas.DataFrame
) -> pandas.DataFrame:
    """Filters dataframe to keep only sources in the cell

    Parameters
    ----------
    cell:
        The cell being analyzed

    iCat:
        Catalog index

    dataframe:
        Input dataframe


    Returns
    -------
    Filtered datasets


    Notes
    -----
    This will optionally deshear the source positions if `matcher.deshear`
    is not None.

    This will add these columns to the output dataframes

    +-----------+-------------------------------------+
    | Column    | Description                         |
    +===========+=====================================+
    | xCell     | X-coordinate in cell frame          |
    +-----------+-------------------------------------+
    | yCell     | Y-coordinate in cell frame          |
    +-----------+-------------------------------------+
    | xPix      | X-coordinate in global WCS frame    |
    +-----------+-------------------------------------+
    | yPix      | Y-coordinate in global WCS frame    |
    +-----------+-------------------------------------+
    | dxShear   | Change in X position when desheared |
    +-----------+-------------------------------------+
    | dyShear   | Change in X position when desheared |
    +-----------+-------------------------------------+

    """

    matcher = cell.matcher

    if TYPE_CHECKING:
        assert isinstance(matcher, ShearMatch)

    filteredIdx = matcher.getCellIndices(dataframe) == cell.idx
    reduced = dataframe[filteredIdx].copy(deep=True)

    xCellOrig = reduced["xCellCoadd"]
    yCellOrig = reduced["yCellCoadd"]
    xPixOrig = reduced["xPix"]
    yPixOrig = reduced["yPix"]

    if matcher.deshear is not None:
        # De-shear in the cell frame to do matching
        dxShear = matcher.deshear * (
            xCellOrig * DESHEAR_COEFFS[iCat][0] + yCellOrig * DESHEAR_COEFFS[iCat][2]
        )
        dyShear = matcher.deshear * (
            xCellOrig * DESHEAR_COEFFS[iCat][1] + yCellOrig * DESHEAR_COEFFS[iCat][3]
        )
        xCell = xCellOrig + dxShear
        yCell = yCellOrig + dyShear
        xPix = xPixOrig + dxShear
        yPix = yPixOrig + dyShear
    else:  # pragma: no cover
        dxShear = np.zeros(len(dataframe))
        dyShear = np.zeros(len(dataframe))
        xCell = xCellOrig
        yCell = yCellOrig
        xPix = xPixOrig
        yPix = yPixOrig

    xCell = (xCell + 100) / matcher.pixelMatchScale
    yCell = (yCell + 100) / matcher.pixelMatchScale
    filteredX = np.bitwise_and(xCell >= 0, xCell < cell.nPix[0])
    filteredY = np.bitwise_and(yCell >= 0, yCell < cell.nPix[1])
    filteredBounds = np.bitwise_and(filteredX, filteredY)
    red = reduced[filteredBounds].copy(deep=True)
    red["xCell"] = xCell[filteredBounds]
    red["yCell"] = yCell[filteredBounds]
    red["xPix"] = xPix[filteredBounds]
    red["yPix"] = yPix[filteredBounds]
    if matcher.deshear is not None:
        red["dxShear"] = dxShear[filteredBounds]
        red["dyShear"] = dyShear[filteredBounds]
    return red



from click.testing import CliRunner, Result
from hpmcm.cli.commands import cli


def check_result(
    result: Result,
) -> None:
    """Check the result of an invocation"""
    if not result.exit_code == 0:
        raise ValueError(f"{result} failed with {result.exit_code} {result.output}")


def test_cli_help() -> None:
    """Make sure the hpmcm --help command works"""
    runner = CliRunner()

    result = runner.invoke(cli, "--help")
    check_result(result)


def test_cli_shear() -> None:
    """Make sure the hpmcm shear --help command works"""
    runner = CliRunner()

    result = runner.invoke(cli, "shear --help")
    check_result(result)


def test_cli_shear_match() -> None:
    """Make sure the hpmcm shear match --help command works"""
    runner = CliRunner()

    result = runner.invoke(cli, "shear match --help")
    check_result(result)



from __future__ import annotations

import numpy as np
from scipy import ndimage
from photutils.segmentation import detect_sources


class Footprint:
    """Wraps the slices returns by `ndimage.find_objects`

    Attributes
    ----------
    image: np.ndaray
        Original image

    sliceX: slice
        Slice of footprint in X

    sliceY: slice
        Slice of footprint in Y

    cutout: np.ndaray
        Image cutout
    """

    def __init__(self, image: np.ndarray, slices: tuple[slice, slice]):
        """Build a Footprint

        Parameters
        ----------
        image:
            Original image (counts map of source positions)

        slice:
            Slices in X and Y
        """
        self.image = image
        self.sliceX = slices[0]
        self.sliceY = slices[1]
        self.cutout = self.image[self.sliceX, self.sliceY]

    def extent(self) -> tuple[int, int, int, int]:
        """Return the extent of the Footprint, for use by `matplotlib`"""
        return (
            self.sliceX.start,
            self.sliceX.stop,
            self.sliceY.start,
            self.sliceY.stop,
        )


class FootprintSet:
    """Wraps Footprints detected by `photutils.segmentation.detect_sources`

    Attributes
    ----------
    image: np.ndarray
        Original image

    footprints: list[Footprint]
        Footprints found in the image

    fpKey: np.ndarray
        Map of footprint associations: -1 -> no association
    """

    def __init__(
        self, image: np.ndarray, fpKey: np.ndarray, footprints: list[Footprint]
    ):
        """Create a FootprintSet

        Parameters
        ----------
        image:
            Original image (counts map of source positions)

        fpKey:
            Map of footprint associations: -1 -> no association

        footprints:
            List of Footprints in this set
        """
        self.image = image
        self.footprints = footprints
        self.fpKey = fpKey

    @classmethod
    def detect(cls, image: np.ndarray) -> FootprintSet:
        """Create a FootprintSet from a countsMap

        Parameters
        ----------
        image:
            Original image (counts map of source positions)

        Returns
        -------
        Newly created FootprintSet
        """
        fpKey = detect_sources(image, 0.5, 1).data
        slices = ndimage.find_objects(fpKey)
        fpKey = fpKey - 1
        footprints: list[Footprint] = []
        for slicePair in slices:
            footprints.append(Footprint(image, slicePair))
        return cls(image, fpKey, footprints)

    def filter(self, buf: int, pixelMatchScale: int = 1) -> FootprintSet:
        """Remove footprints outside the central region of cell"""
        if buf == 0:
            return self

        # The easiest way to do this is to mask out the outer
        # region in the footprint key
        nExclude = np.floor(buf/pixelMatchScale).astype(int)
        nX, nY = self.fpKey.shape
        self.fpKey[0:nExclude] = -1
        self.fpKey[nX-nExclude:nX] = -1
        self.fpKey[:,0:nExclude] = -1
        self.fpKey[:,nY-nExclude:nY] = -1
        return self



from __future__ import annotations

import pickle

import matplotlib.pyplot as plt
import numpy as np
import pandas
from matplotlib.figure import Figure

SHEAR_NAMES = ["ns", "2p", "2m", "1p", "1m"]


class ShearHistogramStats:
    """Simple class to store stats about a histogram

    Attributes
    ----------
    w: float
        Sum of weights

    mean: float
        Histogram mean

    std: float
        Histogram standard deviation

    error: float
        Error on histogram mean

    invVar: float
        Inverse Variance
    """

    def __init__(
        self,
        weights: np.ndarray,
        binCenters: np.ndarray,
    ):
        """Compute the stats from a histogram"""
        self.w = np.sum(weights)
        self.mean = np.sum(weights * binCenters) / self.w
        deltas = binCenters - self.mean
        var = np.sum(weights * deltas * deltas) / self.w
        self.std = np.sqrt(var)
        self.error = self.std / np.sqrt(self.w)
        self.invVar = 1.0 / (self.error * self.error)


class ShearProfileHistogramStats:
    """Simple class to store stats about a 2d histogram

    Attributes
    ----------
    w: np.array
        Sum of weights

    mean: np.array
        Histogram mean

    std: np.array
        Histogram standard deviation

    error: np.array
        Error on histogram mean

    invVar: np.array
        Inverse Variance
    """

    def __init__(
        self,
        hist_2d: tuple[np.ndarray, np.ndarray, np.ndarray],
    ):
        """Compute the stats from a histogram"""
        weights = hist_2d[0]
        yBins = hist_2d[2]
        yBinCenters = 0.5 * (yBins[0:-1] + yBins[1:])

        self.w = np.sum(weights, axis=1)
        self.mean = np.sum(weights * yBinCenters, axis=1) / self.w
        deltas = yBinCenters - np.expand_dims(self.mean, -1)
        var = np.sum(weights * deltas * deltas, axis=1) / self.w
        self.std = np.sqrt(var)
        self.error = self.std / np.sqrt(self.w)
        self.invVar = 1.0 / (self.error * self.error)


class ShearHistograms:
    """Simple class to store histogram relating to shear calibration

    {type} is the matching type, one of "good", "bad", "all"

    {i}, {j} are the components of the shear: 1, 2

    Attributes
    ----------
    binEdges: np.ndarray
        Bin edges for all histograms

    binCenters: np.ndarray
        Bin centers for all histograms

    central: slice
        Slice to select central region of histogram

    centralEdges: slice
        Slice to select edges for central region of histogram

    good_delta_g_{i}_{j}: np.ndarray
        Histogram of g_{i}_{j}p - g_{i}_{j}m for all well-matched objects

    {type}_g_{i}_{cat}: np.ndarray
       Histogram of all g_{i} value of all objects of {type} in {cat}
    """

    def __init__(
        self,
        good: pandas.DataFrame,
        bad: pandas.DataFrame,
        catType: str,
    ):

        if catType != "pgauss":
            self.binEdges = np.linspace(-1, 1, 2001)
            self.central = slice(800, 1200)
            self.centralEdges = slice(800, 1201)
        else:
            self.binEdges = np.linspace(-10, 10, 20001)
            self.central = slice(9800, 10200)
            self.centralEdges = slice(9800, 10201)

        self.binCenters = (self.binEdges[1:] + self.binEdges[0:-1]) / 2.0

        self.good_delta_g_1_1 = np.histogram(good.delta_g_1_1, bins=self.binEdges)[0]
        self.good_delta_g_2_2 = np.histogram(good.delta_g_2_2, bins=self.binEdges)[0]
        self.good_delta_g_1_2 = np.histogram(good.delta_g_1_2, bins=self.binEdges)[0]
        self.good_delta_g_2_1 = np.histogram(good.delta_g_2_1, bins=self.binEdges)[0]

        self.good_g_2_2p = np.histogram(
            good.g_2_2p, weights=good.n_2p, bins=self.binEdges
        )[0]
        self.good_g_2_2m = np.histogram(
            -1 * good.g_2_2m, weights=good.n_2m, bins=self.binEdges
        )[0]
        self.good_g_1_1p = np.histogram(
            good.g_1_1p, weights=good.n_1p, bins=self.binEdges
        )[0]
        self.good_g_1_1m = np.histogram(
            -1 * good.g_1_1m, weights=good.n_1m, bins=self.binEdges
        )[0]

        self.good_g_2_1p = np.histogram(
            good.g_2_1p, weights=good.n_1p, bins=self.binEdges
        )[0]
        self.good_g_2_1m = np.histogram(
            -1 * good.g_2_1m, weights=good.n_1m, bins=self.binEdges
        )[0]
        self.good_g_1_2p = np.histogram(
            good.g_1_2p, weights=good.n_2p, bins=self.binEdges
        )[0]
        self.good_g_1_2m = np.histogram(
            -1 * good.g_1_2m, weights=good.n_2m, bins=self.binEdges
        )[0]

        self.bad_g_2_2p = np.histogram(
            bad.g_2_2p, weights=bad.n_2p, bins=self.binEdges
        )[0]
        self.bad_g_2_2m = np.histogram(
            -1 * bad.g_2_2m, weights=bad.n_2m, bins=self.binEdges
        )[0]
        self.bad_g_1_1p = np.histogram(
            bad.g_1_1p, weights=bad.n_1p, bins=self.binEdges
        )[0]
        self.bad_g_1_1m = np.histogram(
            -1 * bad.g_1_1m, weights=bad.n_1m, bins=self.binEdges
        )[0]

        self.bad_g_2_1p = np.histogram(
            bad.g_2_1p, weights=bad.n_1p, bins=self.binEdges
        )[0]
        self.bad_g_2_1m = np.histogram(
            -1 * bad.g_2_1m, weights=bad.n_1m, bins=self.binEdges
        )[0]
        self.bad_g_1_2p = np.histogram(
            bad.g_1_2p, weights=bad.n_2p, bins=self.binEdges
        )[0]
        self.bad_g_1_2m = np.histogram(
            -1 * bad.g_1_2m, weights=bad.n_2m, bins=self.binEdges
        )[0]

        self.all_g_2_2p = self.bad_g_2_2p + self.good_g_2_2p
        self.all_g_2_2m = self.bad_g_2_2m + self.good_g_2_2m
        self.all_g_1_1p = self.bad_g_1_1p + self.good_g_1_1p
        self.all_g_1_1m = self.bad_g_1_1m + self.good_g_1_1m

        self.all_g_2_1p = self.bad_g_2_1p + self.good_g_2_1p
        self.all_g_2_1m = self.bad_g_2_1m + self.good_g_2_1m
        self.all_g_1_2p = self.bad_g_1_2p + self.good_g_1_2p
        self.all_g_1_2m = self.bad_g_1_2m + self.good_g_1_2m

    def plotMetacalib(
        self,
        stats_g_1: ShearHistogramStats | None = None,
        stats_g_2: ShearHistogramStats | None = None,
        shear: float = 0.01,
        *,
        useCentral: bool = True,
    ) -> Figure:
        """Plot delta_g_1_1 and delta_g_2_2 for fully matched objects"""

        if stats_g_1 is None:
            stats_g_1 = ShearHistogramStats(self.good_delta_g_1_1, self.binCenters)
        if stats_g_2 is None:
            stats_g_2 = ShearHistogramStats(self.good_delta_g_2_1, self.binCenters)
        assert stats_g_1 is not None
        assert stats_g_2 is not None

        fig, axes = plt.subplots()
        if useCentral:
            valSlice = self.central
            edgeSlice = self.centralEdges
        else:
            valSlice = slice(0, len(self.good_delta_g_1_1))
            edgeSlice = slice(0, len(self.binEdges))

        axes.stairs(
            self.good_delta_g_1_1[valSlice],
            self.binEdges[edgeSlice],
            label=f"R_11: {stats_g_1.mean/shear:.4f} +- {stats_g_1.error/shear:.4f}",
        )
        axes.stairs(
            self.good_delta_g_2_2[valSlice],
            self.binEdges[edgeSlice],
            label=f"R_22: {stats_g_2.mean/shear:.4f} +- {stats_g_2.error/shear:.4f}",
        )
        axes.axvline(x=stats_g_1.mean, color="blue", linestyle="-", linewidth=2)
        axes.axvline(x=stats_g_2.mean, color="orange", linestyle="-", linewidth=2)
        axes.axvline(
            x=stats_g_1.mean + stats_g_1.error,
            color="blue",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_2.mean + stats_g_2.error,
            color="orange",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_1.mean - stats_g_1.error,
            color="blue",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_2.mean - stats_g_2.error,
            color="orange",
            linestyle="--",
            linewidth=2,
        )
        axes.legend(loc="upper right")
        axes.set_xlabel(r"$\delta g$")
        axes.set_ylabel("Counts")
        fig.tight_layout()
        return fig

    def plotMetaDetect(
        self,
        hist1p: np.ndarray,
        hist1m: np.ndarray,
        hist2p: np.ndarray,
        hist2m: np.ndarray,
        stats_g_1: ShearHistogramStats | None = None,
        stats_g_2: ShearHistogramStats | None = None,
        shear: float = 0.01,
        title: str = "",
        *,
        useCentral: bool = True,
    ) -> Figure:
        """Plot hist1p - hist1m and hist2p - hist2m for all objects of a particular type"""

        if stats_g_1 is None:
            stats_g_1 = ShearHistogramStats(hist1p + hist1m, self.binCenters)
        if stats_g_2 is None:
            stats_g_2 = ShearHistogramStats(hist2p + hist2m, self.binCenters)
        assert stats_g_1 is not None
        assert stats_g_2 is not None

        fig, axes = plt.subplots()
        if useCentral:
            valSlice = self.central
            edgeSlice = self.centralEdges
        else:
            valSlice = slice(0, len(hist1p))
            edgeSlice = slice(0, len(hist1p) + 1)

        axes.stairs(
            (hist1p + hist1m)[valSlice],
            self.binEdges[edgeSlice],
            label=f"R_11: {2*stats_g_1.mean/shear:.4f} +- {2*stats_g_1.error/shear:.4f}",
        )
        axes.stairs(
            (hist2p + hist2m)[valSlice],
            self.binEdges[edgeSlice],
            label=f"R_22: {2*stats_g_2.mean/shear:.4f} +- {2*stats_g_2.error/shear:.4f}",
        )
        axes.set_xlabel("g")
        axes.set_ylabel("Counts")
        axes.axvline(x=stats_g_1.mean, color="blue", linestyle="-", linewidth=2)
        axes.axvline(x=stats_g_2.mean, color="orange", linestyle="-", linewidth=2)
        axes.axvline(
            x=stats_g_1.mean + stats_g_1.error,
            color="blue",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_2.mean + stats_g_2.error,
            color="orange",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_1.mean - stats_g_1.error,
            color="blue",
            linestyle="--",
            linewidth=2,
        )
        axes.axvline(
            x=stats_g_2.mean - stats_g_2.error,
            color="orange",
            linestyle="--",
            linewidth=2,
        )

        axes.legend(loc="upper right")
        axes.set_title(title)
        fig.tight_layout()
        return fig

    def plotMetaDetectGood(
        self,
        stats_g_1: ShearHistogramStats | None = None,
        stats_g_2: ShearHistogramStats | None = None,
        shear: float = 0.01,
        *,
        useCentral: bool = True,
    ) -> Figure:
        """Plot hist1p - hist1m and hist2p - hist2m for all fully matched objects"""

        return self.plotMetaDetect(
            self.good_g_1_1p,
            self.good_g_1_1m,
            self.good_g_2_2p,
            self.good_g_2_2m,
            stats_g_1,
            stats_g_2,
            shear=shear,
            title="Good Matches",
            useCentral=useCentral,
        )

    def plotMetaDetectAll(
        self,
        stats_g_1: ShearHistogramStats | None = None,
        stats_g_2: ShearHistogramStats | None = None,
        shear: float = 0.01,
        *,
        useCentral: bool = True,
    ) -> Figure:
        """Plot hist1p - hist1m and hist2p - hist2m for all objects"""

        return self.plotMetaDetect(
            self.all_g_1_1p,
            self.all_g_1_1m,
            self.all_g_2_2p,
            self.all_g_2_2m,
            stats_g_1,
            stats_g_2,
            shear=shear,
            title="All Clusters",
            useCentral=useCentral,
        )

    def plotMetaDetectBad(
        self,
        stats_g_1: ShearHistogramStats | None = None,
        stats_g_2: ShearHistogramStats | None = None,
        shear: float = 0.01,
        *,
        useCentral: bool = True,
    ) -> Figure:
        """Plot hist1p - hist1m and hist2p - hist2m for non-fully matched objects"""
        return self.plotMetaDetect(
            self.bad_g_1_1p,
            self.bad_g_1_1m,
            self.bad_g_2_2p,
            self.bad_g_2_2m,
            stats_g_1,
            stats_g_2,
            shear=shear,
            title="Bad Matches",
            useCentral=useCentral,
        )


class ShearStats:
    """Simple class to store shear statisitics

    {type} is the matching type, one of "good", "bad", "all"

    {i}, {j} are the components of the shear: 1, 2

    Attributes
    ----------
    delta_g_{i}_{i}: ShearHistogramStats
        Stats for g_{i}_{j}p - g_{i}_{j}m for fully matched objects

    {type}_g_{i}_{j}: ShearHistogramStats
        Stats for g_{i}_{j} for objects of {type}
    """

    def __init__(
        self,
        hists: ShearHistograms,
    ):

        self.delta_g_1_1 = ShearHistogramStats(hists.good_delta_g_1_1, hists.binCenters)
        self.delta_g_2_2 = ShearHistogramStats(hists.good_delta_g_2_2, hists.binCenters)

        self.delta_g_1_2 = ShearHistogramStats(hists.good_delta_g_1_2, hists.binCenters)
        self.delta_g_2_1 = ShearHistogramStats(hists.good_delta_g_2_1, hists.binCenters)

        self.good_g_1_1 = ShearHistogramStats(
            hists.good_g_1_1p + hists.good_g_1_1m, hists.binCenters
        )
        self.good_g_2_2 = ShearHistogramStats(
            hists.good_g_2_2p + hists.good_g_2_2m, hists.binCenters
        )
        self.good_g_1_2 = ShearHistogramStats(
            hists.good_g_1_2p + hists.good_g_1_2m, hists.binCenters
        )
        self.good_g_2_1 = ShearHistogramStats(
            hists.good_g_2_1p + hists.good_g_2_1m, hists.binCenters
        )

        self.bad_g_1_1 = ShearHistogramStats(
            hists.bad_g_1_1p + hists.bad_g_1_1m, hists.binCenters
        )
        self.bad_g_2_2 = ShearHistogramStats(
            hists.bad_g_2_2p + hists.bad_g_2_2m, hists.binCenters
        )
        self.bad_g_1_2 = ShearHistogramStats(
            hists.bad_g_1_2p + hists.bad_g_1_2m, hists.binCenters
        )
        self.bad_g_2_1 = ShearHistogramStats(
            hists.bad_g_2_1p + hists.bad_g_2_1m, hists.binCenters
        )

        self.all_g_1_1 = ShearHistogramStats(
            hists.all_g_1_1p + hists.all_g_1_1m, hists.binCenters
        )
        self.all_g_2_2 = ShearHistogramStats(
            hists.all_g_2_2p + hists.all_g_2_2m, hists.binCenters
        )
        self.all_g_1_2 = ShearHistogramStats(
            hists.all_g_1_2p + hists.all_g_1_2m, hists.binCenters
        )
        self.all_g_2_1 = ShearHistogramStats(
            hists.all_g_2_1p + hists.all_g_2_1m, hists.binCenters
        )


class ShearData:
    """Collection of shear related data for a single catalog

    Attritubes
    ----------
    shear: float
        Applied shear

    catType: str
        Catalog type

    tract: int
        Tract

    nObjects: int
        Number of objects in catalog

    nInCell: int
        Nubmer of objects in the central region of cell

    nUsed: int
        Number of objects passing SNR cut and in central region of cell

    nGood: int
        Number of fully matched objects

    nBad: int
        Number on non-fully matched objects

    nAll: int
        Number of fully and non-fully matched objects

    effic: float
        Efficiency to fully match objects

    efficErr: float
        Error on efficiency to fully match objects

    hists: ShearHistograms
        Histograms of shear data

    stats: ShearStats
        Summary statistics
    """

    def __init__(
        self,
        shearTable: pandas.DataFrame,
        statsTable: pandas.DataFrame,
        shear: float,
        catType: str,
        tract: int,
        snrCut: float = 7.5,
    ):
        shearTable["idx"] = np.arange(len(shearTable))
        statsTable["idx"] = np.arange(len(statsTable))
        merged = shearTable.merge(statsTable, on="idx")

        self.shear = shear
        self.catType = catType
        self.tract = tract

        in_cell_mask = np.bitwise_and(
            np.fabs(merged.xCent - 100) < 75,
            np.fabs(merged.yCent - 100) < 75,
        )
        in_cell = merged[in_cell_mask]
        bright_mask = in_cell.SNR > snrCut

        used = in_cell[bright_mask]
        good_mask = used.good
        good = used[good_mask]
        bad = used[~good_mask]

        self.nObjects = len(merged)
        self.nInCell = len(in_cell)
        self.nUsed = len(used)
        self.nGood = len(good)
        self.nBad = len(bad)
        self.nAll = self.nGood + self.nBad
        self.effic = self.nGood / self.nAll
        self.efficErr = float(np.sqrt(self.effic * (1.0 - self.effic) / self.nAll))

        print("")
        print(f"Report: {catType} {shear:.4f} {tract}")
        print("All Objects:                               ", self.nObjects)
        print("InCell                                     ", self.nInCell)
        print("Used                                       ", self.nUsed)
        print("Good                                       ", self.nGood)
        print("Bad                                        ", self.nBad)
        print(
            "Efficiency                                 ",
            f"{self.effic:.4f} +- {self.efficErr:.4f}",
        )
        self.hists = ShearHistograms(good, bad, self.catType)
        self.stats = ShearStats(self.hists)

    def makePlots(self, *, useCentral: bool = True) -> dict[str, Figure]:
        """Make the standard plots"""
        plotDict = dict(
            metacalib=self.hists.plotMetacalib(
                self.stats.delta_g_1_1,
                self.stats.delta_g_2_2,
                shear=self.shear,
                useCentral=useCentral,
            ),
            metadetect_good=self.hists.plotMetaDetectGood(
                self.stats.good_g_1_1,
                self.stats.good_g_2_2,
                shear=self.shear,
                useCentral=useCentral,
            ),
            metadetect_bad=self.hists.plotMetaDetectBad(
                self.stats.bad_g_1_1,
                self.stats.bad_g_2_2,
                shear=self.shear,
                useCentral=useCentral,
            ),
            metadetect_all=self.hists.plotMetaDetectAll(
                self.stats.all_g_1_1,
                self.stats.all_g_2_2,
                shear=self.shear,
                useCentral=useCentral,
            ),
        )
        return plotDict

    def toDict(self) -> dict:
        """Convert self to a dict"""
        outDict = {}
        outDict["shear"] = self.shear
        outDict["n_objects"] = self.nObjects
        outDict["n_in_cell"] = self.nInCell
        outDict["n_used"] = self.nUsed
        outDict["n_good"] = self.nGood
        outDict["n_bad"] = self.nBad
        outDict["efficiency"] = self.effic
        outDict["efficiency_err"] = self.efficErr

        prefixes = [
            "mc_delta_g_1_1",
            "mc_delta_g_2_2",
            "mc_delta_g_1_2",
            "mc_delta_g_2_1",
            "md_all_g_1_1",
            "md_all_g_2_2",
            "md_all_g_1_2",
            "md_all_g_2_1",
            "md_good_g_1_1",
            "md_good_g_2_2",
            "md_good_g_1_2",
            "md_good_g_2_1",
            "md_bad_g_1_1",
            "md_bad_g_2_2",
            "md_bad_g_1_2",
            "md_bad_g_2_1",
        ]
        all_stats = [
            self.stats.delta_g_1_1,
            self.stats.delta_g_2_2,
            self.stats.delta_g_1_2,
            self.stats.delta_g_2_1,
            self.stats.all_g_1_1,
            self.stats.all_g_2_2,
            self.stats.all_g_1_2,
            self.stats.all_g_2_1,
            self.stats.good_g_1_1,
            self.stats.good_g_2_2,
            self.stats.good_g_1_2,
            self.stats.good_g_2_1,
            self.stats.bad_g_1_1,
            self.stats.bad_g_2_2,
            self.stats.bad_g_1_2,
            self.stats.bad_g_2_1,
        ]

        for prefix_, stats_ in zip(prefixes, all_stats):
            outDict[f"{prefix_}"] = stats_.mean
            outDict[f"{prefix_}_std"] = stats_.std
            outDict[f"{prefix_}_err"] = stats_.error
            outDict[f"{prefix_}_invVar"] = stats_.invVar
        return outDict

    def save(self, filepath: str) -> None:
        """Save to a pickle file"""
        with open(filepath, "wb") as fOut:
            pickle.dump(self, fOut)

    @classmethod
    def load(cls, filepath: str) -> ShearData:
        """Load from a pickle file"""
        with open(filepath, "rb") as fIn:
            loaded = pickle.load(fIn)
        return loaded

    def savefigs(self, outputFileBase: str) -> None:
        """Save all the figures"""
        plots = self.makePlots()
        for key, plot_ in plots.items():
            plot_.savefig(f"{outputFileBase}_{key}.png")



from __future__ import annotations

import sys
from collections import OrderedDict
from collections.abc import Iterable
from typing import Any

import numpy as np
import pandas
import pyarrow.parquet as pq

from . import input_tables, output_tables
from .cell import CellData
from .cluster import ClusterData
from .object import ObjectData


class Match:
    """Class to do N-way matching

    Uses a provided WCS to define a Skymap that covers the full region
    begin matched.

    Uses that WCS to assign pixel locations to all sources in the input catalogs

    Iterates over cells and does source clustering in each cell
    using Footprint detection on a Skymap of source counts per pixel.

    Assigns each input source to a cluster.

    At that stage the clusters are not the final product as they can include
    more than one soruce from a given catalog.

    Loops over clusters and processes each cluster to resolve confusion.

    If there is not a unqiue source per-catalog redo the clustering with
    half-size pixels to try to split the cluster (down to minimum pixel scale)

    Attributes
    ----------
    pixSize : float
        Pixel size in arcseconds

    nPixSide: int
        Number of pixels in the match region

    cellSize: int
        Number of pixels in a Cell

    cellBuffer: int
        Number of overlapping pixel in a Cell

    cellMaxObject: int
        Max number of objects in a cell, used to make unique IDs

    maxSubDivision: int
        Maximum number of cell sub-divisions

    pixelR2Cut: float
        Distance cut for Object membership, in pixels**2

    nCell: np.ndarray
        Number of cells in match region

    fullData: list[DataFrame]
        Full input DataFrames

    redData : list[DataFrame]
        Reduced DataFrames with only the columns needed for matching

    cellDict : OrderedDict[int, CellData]
        Dictionary providing access to cell data

    Notes
    -----
    This expectes a list of parquet files with pandas DataFrames.
    The expected columns depend on which sub-class of `Match` is being used.

    Four output tables are produced:

    +----------------+---------------------------------------------------+
    | Key            | Class                                             |
    +================+===================================================+
    | _cluster_assoc | :py:class:`hpmcm.output_tables.ClusterAssocTable` |
    +----------------+---------------------------------------------------+
    | _cluster_stats | :py:class:`hpmcm.output_tables.ClusterStatsTable` |
    +----------------+---------------------------------------------------+
    | _object_assoc  | :py:class:`hpmcm.output_tables.ObjectAssocTable`  |
    +----------------+---------------------------------------------------+
    | _object_stats  | :py:class:`hpmcm.output_tables.ObjectStatsTable`  |
    +----------------+---------------------------------------------------+

    """

    inputTableClass: type = input_tables.SourceTable
    extraCols: list[str] = []

    def __init__(
        self,
        **kwargs: Any,
    ):
        self.pixSize = kwargs["pixelSize"]
        self.nPixSide = kwargs["nPixels"]
        self.cellSize: int = kwargs.get("cellSize", 1000)
        self.cellBuffer: int = kwargs.get("cellBuffer", 10)
        self.cellMaxObject: int = kwargs.get("cellMaxObject", 100000)
        self.maxSubDivision: int = kwargs.get("maxSubDivision", 3)
        self.pixelR2Cut: float = kwargs.get("pixelR2Cut", 1.0)
        self.nCell: np.ndarray = np.ceil(self.nPixSide / self.cellSize)

        self.fullData: OrderedDict[int, pandas.DataFrame] = OrderedDict()
        self.redData: OrderedDict[int, pandas.DataFrame] = OrderedDict()
        self.cellDict: OrderedDict[int, CellData] = OrderedDict()

    def pixToArcsec(self) -> float:
        """Convert pixel size (in degrees) to arcseconds"""
        return 3600.0 * self.pixSize

    def pixToWorld(
        self,
        xPix: np.ndarray,
        yPix: np.ndarray,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Convert local coords in pixels to world coordinates (RA, DEC)"""
        return np.repeat(np.nan, len(xPix)), np.repeat(np.nan, len(yPix))

    def getCellIdx(
        self,
        ix: int,
        iy: int,
    ) -> int:
        """Get the Index to use for a given cell"""
        return int(self.nCell[1] * ix + iy)

    def getIdOffset(
        self,
        ix: int,
        iy: int,
    ) -> int:
        """Get the ID offset to use for a given cell"""
        cellIdx = self.getCellIdx(ix, iy)
        return int(self.cellMaxObject * cellIdx)

    def reduceData(
        self,
        inputFiles: list[str],
        catalogId: list[int],
    ) -> None:
        """Read input files and filter out only the columns we need

        Each input file should have an associated catalogId.
        This is used to test if we have more than one-source
        per input catalog.

        If the inputs files have a pre-defined ID associated with them
        that can be used.   Otherwise it is fine just to give a range from
        0 to nInputs.
        """
        for fName, cid in zip(inputFiles, catalogId):
            self.fullData[cid] = self._readDataFrame(fName)
            self.redData[cid] = self.reduceDataFrame(self.fullData[cid])
            self.fullData[cid].set_index("id", inplace=True)

    def _buildCellData(
        self,
        idOffset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
    ) -> CellData:
        return CellData(self, idOffset, corner, size, idx, self.cellBuffer)

    def analyzeCell(
        self,
        ix: int,
        iy: int,
        fullData: bool = False,
    ) -> dict | None:
        """Analyze a single cell

        Parameters
        ----------
        ix:
            Cell index in x-coord

        iy:
            Cell index in y-coord

        Returns
        -------
        Output of cell analysis


        Notes
        -----
        cellData : CellData : The analysis data for the Cell

        image : np.ndarray : Image of cell source counts map

        countsMap : np.ndarray : Numpy array with cell source counts

        clusters : FootprintSet : Clusters as dectected by finding FootprintSet on source counts map

        clusterKey : np.ndarray : Map of cell with pixels filled with index of associated Footprints

        Notes
        -----
        If fullData is False, only cellData will be returned
        """
        iCell = self.getCellIdx(ix, iy)
        cellStep = np.array([self.cellSize, self.cellSize])
        corner = np.array([ix - 1, iy - 1]) * cellStep
        idOffset = self.getIdOffset(ix, iy)
        cellData = self._buildCellData(idOffset, corner, cellStep, iCell)
        cellData.reduceData(list(self.redData.values()))
        oDict = cellData.analyze(pixelR2Cut=self.pixelR2Cut)
        if cellData.nObjects >= self.cellMaxObject:  # pragma: no cover
            print("Too many object in a cell", cellData.nObjects, self.cellMaxObject)

        self.cellDict[iCell] = cellData
        if oDict is None:
            return None
        if fullData:  # pragma: no cover
            oDict["cellData"] = cellData
            return oDict

        return dict(cellData=cellData)

    def analysisLoop(
        self, xRange: Iterable | None = None, yRange: Iterable | None = None
    ) -> None:
        """Does matching for all cells.

        This stores the results, but does not write or return them.

        Parameters
        ----------
        xRange:
            Range of cells to analysze in X.  None -> Entire range.

        yRange:
            Range of cells to analysis in Y.  None -> Entire range.
        """
        self.cellDict.clear()

        if xRange is None:
            xRange = range(int(self.nCell[0]))
        if yRange is None:
            yRange = range(int(self.nCell[1]))

        for ix in xRange:
            for iy in yRange:
                odict = self.analyzeCell(ix, iy)
                if odict is None:
                    continue
            if ix == 0:
                pass
            elif ix % 10 == 0:
                sys.stdout.write(f" {ix}!\n")
                sys.stdout.flush()
            else:
                sys.stdout.write(".")
                sys.stdout.flush()

        sys.stdout.write(" Done!\n")
        sys.stdout.flush()

    def extractStats(self) -> list[pandas.DataFrame]:
        """Extracts cluster statisistics

        Returns
        -------
        DataFrames with matching info,
        :py:class:`hpmcm.output_tables.ClusterAssocTable`,
        :py:class:`hpmcm.output_tables.ObjectAssocTable`.
        :py:class:`hpmcm.output_tables.ClusterStatsTable`.
        :py:class:`hpmcm.output_tables.ObjectStatsTable`.

        """
        clusterAssocTables = []
        objectAssocTables = []
        clusterStatsTables = []
        objectStatsTables = []

        for ix in range(int(self.nCell[0])):
            for iy in range(int(self.nCell[1])):
                iCell = self.getCellIdx(ix, iy)
                if iCell not in self.cellDict:
                    continue
                cellData = self.cellDict[iCell]
                clusterAssocTables.append(
                    output_tables.ClusterAssocTable.buildFromCellData(cellData).data,
                )
                objectAssocTables.append(
                    output_tables.ObjectAssocTable.buildFromCellData(cellData).data,
                )
                clusterStatsTables.append(
                    output_tables.ClusterStatsTable.buildFromCellData(cellData).data,
                )
                objectStatsTables.append(
                    output_tables.ObjectStatsTable.buildFromCellData(cellData).data,
                )
        return [
            pandas.concat(clusterAssocTables),
            pandas.concat(objectAssocTables),
            pandas.concat(clusterStatsTables),
            pandas.concat(objectStatsTables),
        ]

    def _readDataFrame(
        self,
        fName: str,
    ) -> pandas.DataFrame:
        """Read a single input file"""
        # FIXME, we want to use this function
        # return self.inputTableClass.read(fName, self.extraCols)
        parq = pq.read_pandas(fName)
        df = parq.to_pandas()
        return df

    def _getPixValues(self, df: pandas.DataFrame) -> tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError()

    def reduceDataFrame(
        self,
        df: pandas.DataFrame,
    ) -> pandas.DataFrame:
        """Reduce a single input DataFrame

        Parameters
        ----------
        df:
            Input data frame

        Returns
        -------
        Reduced DataFrame
        """
        raise NotImplementedError()

    def getCluster(self, iK: tuple[int, int]) -> ClusterData:
        """Get a particular cluster

        Parameters
        ----------
        iK:
            CellId, ClusterId

        Returns
        -------
        Requested cluster
        """
        cellData = self.cellDict[iK[0]]
        cluster = cellData.clusterDict[iK[1]]
        return cluster

    def getObject(self, iK: tuple[int, int]) -> ObjectData:
        """Get a particular object

        Parameters
        ----------
        iK:
            CellId, ObjectId

        Returns
        -------
        Requested object
        """
        cellData = self.cellDict[iK[0]]
        theObj = cellData.objectDict[iK[1]]
        return theObj



import glob
import os

import numpy as np

import hpmcm

DATADIR = "examples/test_data"  # Input data directory
SHEAR_ST = "0p01"  # Applied shear as a string
SHEAR = 0.01  # Decimal version of applied shear
CATALOG_TYPE = "wmom"  # which object characterization to use
TRACT = 10463  # which tract to study

BASEFILE = os.path.join(DATADIR, f"shear_{CATALOG_TYPE}_{SHEAR_ST}_match_{TRACT}")

PIXEL_R2CUT = 4.0  # Cut at distance**2 = 4 pixels
PIXEL_MATCH_SCALE = 1  # Use pixel scale to do matching
SNR_CUT = 10.0  # Cut on signal-to-noise


def testShearMatch(setup_data: int) -> None:
    """Run ShearMatch on 400 cells"""

    assert setup_data == 0

    # get the data
    source_tablesfiles = sorted(
        glob.glob(
            os.path.join(DATADIR, f"shear_{CATALOG_TYPE}_{SHEAR_ST}_uncleaned_{TRACT}_*.pq")
        )
    )
    source_tablesfiles.reverse()
    catalog_ids = np.arange(len(source_tablesfiles))

    
    # Create matcher
    matcher = hpmcm.ShearMatch.createShearMatch(
        pixelR2Cut=PIXEL_R2CUT, pixelMatchScale=PIXEL_MATCH_SCALE, deshear=-1 * SHEAR
    )

    # Reduce the input data
    matcher.reduceData(source_tablesfiles, catalog_ids)

    # Make sure it got the right number of cells
    assert matcher.nCell[0] == 200
    assert matcher.nCell[1] == 200

    # Define the range of cells to run over
    xRange = range(50, 70)
    yRange = range(170, 190)

    # Run the analysis
    matcher.analysisLoop(xRange, yRange)

    # Extract some data
    stats = matcher.extractStats()
    shear_stats = matcher.extractShearStats()

    assert stats is not None
    assert shear_stats is not None
    obj_shear = shear_stats[1]
    assert obj_shear is not None

    # Test the classification codes
    hpmcm.classify.printSummaryStats(matcher)

    objLists = hpmcm.classify.classifyObjects(matcher, SNRCut=10.0)
    hpmcm.classify.printObjectTypes(objLists)

    clusterLists = hpmcm.classify.classifyClusters(matcher, SNRCut=10.0)
    hpmcm.classify.printClusterTypes(clusterLists)

    # Make sure the match efficiency is high
    n_good = len(objLists["ideal"])
    bad_list = [
        "edge_mixed",
        "edge_missing",
        "edge_extra",
        "orphan",
        "missing",
        "two_missing",
        "many_missing",
        "extra",
        "caught",
    ]
    n_bad = np.sum([len(objLists[x]) for x in bad_list])
    effic = n_good / (n_good + n_bad)
    assert effic > 0.95

    # Get a particular cell and rerun the analysis to test visualization
    # and classification functions
    cellIdx = matcher.getCellIdx(50, 170)
    cell = matcher.cellDict[cellIdx]
    od = cell.analyze(None, 4)
    cluster = matcher.getCluster(clusterLists["ideal"][0])
    obj = matcher.getObject(objLists["ideal"][0])

    assert len(cluster.xCluster)
    assert len(cluster.yCluster)
    assert len(cluster.xPix)
    assert len(cluster.yPix)

    assert len(obj.xCluster)
    assert len(obj.yCluster)
    assert len(obj.xPix)
    assert len(obj.yPix)

    hpmcm.viz_utils.showShearObjs(matcher, clusterLists["ideal"][0])
    hpmcm.viz_utils.showShearObj(matcher, objLists["ideal"][0])
    hpmcm.viz_utils.showCluster(od["image"], cluster, cell)
    hpmcm.viz_utils.showObjects(od["image"], cluster, cell)
    hpmcm.viz_utils.showObjectsV2(od["image"], cluster, cell)


def testShearReport(setup_data: int) -> None:
    """Run the shearReport function"""
    assert setup_data == 0

    output_file = os.path.join(DATADIR, "dummy")
    hpmcm.shear_utils.shearReport(
        BASEFILE,
        output_file,
        SHEAR,
        catType=CATALOG_TYPE,
        tract=TRACT,
        snrCut=SNR_CUT,
    )



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import match_utils, shear_utils, utils
from .cluster import ClusterData, ShearClusterData
from .footprint import Footprint, FootprintSet
from .object import ObjectData, ShearObjectData

if TYPE_CHECKING:
    from .match import Match
    from .shear_match import ShearMatch


class CellData:
    """Class to store analyze data for a cell

    Includes cell boundries, reduced data tables
    and clustering results

    Does not store sky maps

    Cells are square sub-regions of the analysis region
    that are extracted from the ranges of pixels in the WCS

    The cell covers corner:corner+size

    The sources are projected into an array that extends `buf` pixels
    beyond the cell

    This uses the FootprintSet to identify pixels which contain
    source, and builds those into clusters

    Attributes
    ----------
    matcher: Match
        Parent Match object

    idOffset: int
        Offset used for the Object and Cluster IDs for this cell

    corner: np.ndarray
        pixX, pixY for corner of cell

    size: np.ndarray
        size of the cell (in pixels)

    idx: int
        Id of the cell

    buf: int
        Number of buffer pixels around the edge of the cell

    minPix: np.ndarray
        Lowest number pixel in center of cell

    maxPix: np.ndarray
        Highest number pixel in center of cell

    nPix: np.ndarray
        Number of pixels in center of cell

    data : list[pandas.DataFrame]
        Reduced dataframes with only sources for this cell

    nSrc : int
        Number of sources in this cell

    footprintIds : list[np.ndarray]
        Matched arrays with the index of the cluster associated to each
        source.  I.e., these could added to the Dataframes as
        additional columns

    clusterDict : OrderedDict[int, ClusterData]
        Dictionary with cluster membership data

    objectDict : OrderedDict[int, ObjectData]
        Dictionary with object membership data

    """

    def __init__(
        self,
        matcher: Match,
        idOffset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
        buf: int = 10,
    ):
        self.matcher: Match = matcher
        # Offset used for the Object and Cluster IDs for this cell
        self.idOffset: int = idOffset
        self.corner: np.ndarray = corner  # pixX, pixY for corner of cell
        self.size: np.ndarray = size  # size of cell
        self.idx: int = idx  # cell index
        self.buf: int = buf
        self.minPix: np.ndarray = corner - buf
        self.maxPix: np.ndarray = corner + size + buf
        self.nPix: np.ndarray = self.maxPix - self.minPix

        self.data: list[pandas.DataFrame] = []
        self.nSrc: int = 0
        self.footprintIds: list[np.ndarray] = []
        self.clusterDict: OrderedDict[int, ClusterData] = OrderedDict()
        self.objectDict: OrderedDict[int, ObjectData] = OrderedDict()

    def reduceData(self, data: list[pandas.DataFrame]) -> None:
        """Pull out only the data needed for this cell"""
        self.data = [self.reduceDataframe(i, val) for i, val in enumerate(data)]
        self.nSrc = int(np.sum([len(df) for df in self.data]))

    @property
    def nClusters(self) -> int:
        """Return the number of clusters in this cell"""
        return len(self.clusterDict)

    @property
    def nObjects(self) -> int:
        """Return the number of objects in this cell"""
        return len(self.objectDict)

    def reduceDataframe(
        self, iCat: int, dataframe: pandas.DataFrame
    ) -> pandas.DataFrame:
        """Filters dataframe to keep only source in the cell"""

        # WCS is defined, use it
        xCell = dataframe["xPix"] - self.minPix[0]
        yCell = dataframe["yPix"] - self.minPix[1]
        filtered = (
            (xCell >= 0)
            & (xCell < self.nPix[0])
            & (yCell >= 0)
            & (yCell < self.nPix[1])
        )
        red = dataframe[filtered].copy(deep=True)
        red["xCell"] = xCell[filtered]
        red["yCell"] = yCell[filtered]
        return red

    def countsMap(self, weightName: str | None = None) -> np.ndarray:
        """Fill a map that counts the number of source per cell"""
        toFill = self._emtpyCountsMaps()
        assert self.data is not None
        for df in self.data:
            toFill += self._singleCatalogCountsMap(df, weightName)
        return toFill

    def buildClusterData(
        self,
        fpSet: FootprintSet,
        pixelR2Cut: float = 4.0,
    ) -> None:
        """Loop through cluster ids and collect sources into
        the ClusterData objects"""
        footprintDict: dict[int, list[tuple[int, int, int]]] = {}
        nMissing = 0
        nFound = 0
        assert self.data is not None
        assert self.footprintIds
        for iCat, (df, footprintIds) in enumerate(zip(self.data, self.footprintIds)):
            for srcIdx, (srcId, footprintId) in enumerate(zip(df["id"], footprintIds)):
                if footprintId < 0:
                    nMissing += 1
                    continue
                if footprintId not in footprintDict:
                    footprintDict[footprintId] = [(iCat, srcId, srcIdx)]
                else:
                    footprintDict[footprintId].append((iCat, srcId, srcIdx))
                nFound += 1
        for footprintId, sources in footprintDict.items():
            footprint = fpSet.footprints[footprintId]
            iCluster = footprintId + self.idOffset
            cluster = self._buildClusterData(iCluster, footprint, np.array(sources).T)
            self.clusterDict[iCluster] = cluster
            match_utils.heirarchicalProcessCluster(cluster, self, pixelR2Cut)

    def analyze(
        self, weightName: str | None = None, pixelR2Cut: float = 2.0
    ) -> dict | None:
        """Analyze this cell

        Note that this returns the counts maps and clustering info,
        which can be helpful for debugging.
        """
        if self.nSrc == 0:
            return None
        countsMap = self.countsMap(weightName)
        oDict = self._getFootprints(countsMap)
        oDict["countsMap"] = countsMap
        assert self.data is not None
        self.footprintIds = self._associateSourcesToFootprints(
            self.data, oDict["footprintKey"]
        )
        self.buildClusterData(oDict["footprints"], pixelR2Cut)
        return oDict

    def addObject(
        self, cluster: ClusterData, mask: np.ndarray | None = None
    ) -> ObjectData:
        """Add an object to this cell

        Parameters
        ----------
        cluster:
            Parent cluster for the object

        mask:
            Mask of which sources in the cluster to include in the object

        Returns
        -------
        Newly created ObjectData
        """
        objectId = self.nObjects + self.idOffset
        newObject = self._newObject(cluster, objectId, mask)
        self.objectDict[objectId] = newObject
        return newObject

    @classmethod
    def _newObject(
        cls, cluster: ClusterData, objectId: int, mask: np.ndarray | None
    ) -> ObjectData:
        return ObjectData(cluster, objectId, mask)

    def _emtpyCountsMaps(self) -> np.ndarray:
        toFill = np.zeros(np.ceil(self.nPix).astype(int))
        return toFill

    def _singleCatalogCountsMap(
        self, df: pandas.DataFrame, weightName: str | None = None
    ) -> np.ndarray:
        return utils.fillCountsMapFromDf(
            df,
            nPix=self.nPix,
            weightName=weightName,
        )

    def _buildClusterData(
        self, iCluster: int, footprint: Footprint, sources: np.ndarray
    ) -> ClusterData:
        return ClusterData(iCluster, footprint, sources)

    def _getFootprints(self, countsMap: np.ndarray) -> dict:
        return utils.getFootprints(countsMap, buf=self.buf)

    def _associateSourcesToFootprints(
        self,
        data: list[pandas.DataFrame],
        clusterKey: np.ndarray,
    ) -> list[np.ndarray]:
        return utils.associateSourcesToFootprints(
            data,
            clusterKey,
        )

    def getRaDec(
        self, xCents: np.ndarray, yCents: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return the RA, DEC of based from pixel coords"""
        return self.matcher.pixToWorld(xCents, yCents)


class ShearCellData(CellData):
    """Subclass of CellData that can compute shear statisitics

    Attributes
    ----------
    pixelMatchScale: int
        Number of pixel merged in the original counts map
    """

    def __init__(
        self,
        matcher: ShearMatch,
        idOffset: int,
        corner: np.ndarray,
        size: np.ndarray,
        idx: int,
        buf: int = 10,
    ):
        CellData.__init__(self, matcher, idOffset, corner, size, idx, buf)
        self.pixelMatchScale = matcher.pixelMatchScale

    def reduceDataframe(
        self, iCat: int, dataframe: pandas.DataFrame
    ) -> pandas.DataFrame:
        """Filters dataframe to keep only source in the cell"""
        return shear_utils.reduceShearDataForCell(self, iCat, dataframe)

    @classmethod
    def _newObject(
        cls, cluster: ClusterData, objectId: int, mask: np.ndarray | None
    ) -> ObjectData:
        return ShearObjectData(cluster, objectId, mask)

    def _emtpyCountsMaps(self) -> np.ndarray:
        pixelMatchScale = self.pixelMatchScale
        toFill = np.zeros(np.ceil(self.nPix / pixelMatchScale).astype(int))
        return toFill

    def _singleCatalogCountsMap(
        self, df: pandas.DataFrame, weightName: str | None = None
    ) -> np.ndarray:
        return utils.fillCountsMapFromDf(
            df,
            nPix=self.nPix,
            weightName=weightName,
            pixelMatchScale=self.pixelMatchScale,
        )

    def _buildClusterData(
        self, iCluster: int, footprint: Footprint, sources: np.ndarray
    ) -> ClusterData:
        return ShearClusterData(
            iCluster, footprint, sources, pixelMatchScale=self.pixelMatchScale
        )

    def _getFootprints(self, countsMap: np.ndarray) -> dict:
        return utils.getFootprints(
            countsMap, buf=0, pixelMatchScale=self.pixelMatchScale
        )

    def _associateSourcesToFootprints(
        self,
        data: list[pandas.DataFrame],
        clusterKey: np.ndarray,
    ) -> list[np.ndarray]:
        return utils.associateSourcesToFootprints(
            data,
            clusterKey,
            pixelMatchScale=self.pixelMatchScale,
        )

    def getRaDec(
        self, xCents: np.ndarray, yCents: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        return np.repeat(np.nan, len(xCents)), np.repeat(np.nan, len(yCents))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas

from . import shear_utils

if TYPE_CHECKING:
    from .cluster import ClusterData


class ObjectData:
    """Small class to define 'Objects', i.e., sets of associated sources

    Attributes
    ----------
    parentCluster : ClusterData
        Parent cluster that this object is build from

    objectId : int
        Unique object identifier

    mask : np.ndarray[bool]
        Mask of which source in parent cluster are in this object

    catIndices: np.ndarray[int]
        Indicies showing which catalog each source belongs to

    nSrc: int
        Number of sources in this object

    nUnique: int
        Number of catalogs contributing to this object

    data: pandas.DataFrame
        Data for the sources in this object

    recurse: int
        Recursion level needed to make this object

    xCent: float
        X Centeroid of this object in global pixel coordinates

    yCent: float
        Y Centeroid of this object in global pixel coordinates

    rmsDist: float
        RMS distance of sources to the centroid

    dist2: np.ndarray
        Distances of sources to the centroid
    """

    def __init__(
        self,
        cluster: ClusterData,
        objectId: int,
        mask: np.ndarray | None,
        recurse: int = 0,
    ):
        """Build from `ClusterData`, an objectId and mask specifying with sources
        in the cluster are part of the object"""
        self.parentCluster: ClusterData = cluster
        self.objectId: int = objectId
        if mask is None:
            self.mask = np.ones((self.parentCluster.nSrc), dtype=bool)
        else:
            self.mask = mask
        self.catIndices: np.ndarray = self.parentCluster.catIndices[self.mask]
        self.nSrc: int = self.catIndices.size
        self.nUnique: int = np.unique(self.catIndices).size
        self.data: pandas.DataFrame | None = None

        self.recurse: int = recurse
        self.xCent: float = np.nan
        self.yCent: float = np.nan
        self.rmsDist: float = np.nan
        self.snrMean: float = np.nan
        self.snrRms: float = np.nan
        self.dist2: np.ndarray = np.array([])
        self.extract()

    @property
    def xCluster(self) -> np.ndarray:
        """Return the x-position of the sources within the footprint"""
        assert self.data is not None
        return self.data.xCluster

    @property
    def yCluster(self) -> np.ndarray:
        """Return the y-position of the sources within the footprint"""
        assert self.data is not None
        return self.data.yCluster

    @property
    def xPix(self) -> np.ndarray:
        """Return the x-position of the sources within the cell"""
        assert self.data is not None
        return self.data.xPix

    @property
    def yPix(self) -> np.ndarray:
        """Return the y-position of the sources within the cell"""
        assert self.data is not None
        return self.data.yPix

    def hasRefCatalog(self, refCatId: int = 0) -> bool:
        """Is there a source from the reference catalog"""
        return refCatId in self.catIndices

    def sourceIds(self) -> np.ndarray:
        """Return the source ids for the sources in the object"""
        return self.parentCluster.srcIds[self.mask]

    def sourceIdxs(self) -> np.ndarray:
        """Return the source indices for the sources in the object"""
        return self.parentCluster.srcIdxs[self.mask]

    def _updateCatIndices(self) -> None:
        self.catIndices = self.parentCluster.catIndices[self.mask]
        self.nSrc = self.catIndices.size
        self.nUnique = np.unique(self.catIndices).size

    def extract(self) -> None:
        """Extract data from parent cluster"""
        assert self.parentCluster.data is not None
        self.data = self.parentCluster.data.iloc[self.mask]
        self._updateCatIndices()


class ShearObjectData(ObjectData):
    """Subclass of ObjectData that can compute shear statisitics"""

    def shearStats(self) -> dict:
        """Return the shear statistics"""
        assert self.data is not None
        return shear_utils.shearStats(self.data)
